#!/usr/bin/env python
# -*- coding: utf-8 -*-
# Filename: prevot1993.py
# Author: "Thomas Weiß"
# Date:   2017-07-25 16:25:48
# Last Modified by:   "Thomas Weiß"
# Last Modified time: 2017-07-26 13:55:20

"""
This module implements the Prevot et al. (1993)
empirical surface backscattering model component

References
----------
Prevot et al. (1992): Estimation the characteristics of vegetation canopy with airborne radar measurements
"""

import numpy as np
import matplotlib.pyplot as plt

from . scatter import SurfaceScatterWaterCloudModel
import pdb

class Prevot93_surface(SurfaceScatterWaterCloudModel):
    def __init__(self, mv, theta, C1, C2, D):
        """
        Parameters
        ----------
        mv : float
            volumetric soil moisture [m**3/m**3]
        C1 : float
            calibration constant [dB]
        C2 : float
            angular sensitivity of the soil signal [dB/°] (related to roughness)
        theta : float
            incidence angle [rad]
        D : float
            sensitivity of the signal to soil moisture
        """
        super(Prevot93_surface, self).__init__(mv, theta, C1=C1, C2=C2, D=D)
        pdb.set_trace()

        # calculate backscatter
        self.C = self.C1 - self.C2 * self.theta
        self.surface = self.C + self.D * self.mv

    def plot(self):
        f = plt.figure()
        ax = f.add_subplot(111)
        t = np.rad2deg(self.theta)
        ax.plot(t, 10.*np.log10(self.surface), color='blue', label='surface component')
        ax.grid()
        # ax.set_ylim(-25.,0.)
        # ax.set_xlim(0.,70.)
        ax.legend()
        ax.set_xlabel('incidence angle [deg]')
        ax.set_ylabel('backscatter [dB]')
        plt.show()
        pdb.set_trace()










import netCDF4 as nc
import datetime as dt
import numpy as np
import warnings
import pandas as pd


class StateVector:
    """Class to hold state vector data for optical and microwave canopy RT models.
    """
    def __init__(self):

        self.date_utc = None
        self.lai = None
        self.can_height = None
        self.leafChl = None
        self.leafWater = None
        self.soil_moisture = None
        self.soilAlbedo = None


def get_state_csv(fname='/data/state_variables/state_example.csv'):
    """Function that returns StateVector instance for a given file
    :param fname: Path to file from which to extract data, files must be csv with columns:
     (date, lai, canopy height, soil moisture)
    :type fname: str
    :return: StateVector instance
    """
    state_dat = np.loadtxt(fname, delimiter=',', dtype='str')
    state_inst = StateVector()
    state_inst.date_utc = [dt.datetime.strptime(dd, "%Y/%m/%d %H:%M") for dd in state_dat[:,0]]
    state_inst.lai = [float(lai) for lai in state_dat[:,1]]  # (m2 m-2)
    state_inst.can_height = [float(can_height) for can_height in state_dat[:,2]]  # (m)
    state_inst.soil_moisture = [float(soil_m) for soil_m in state_dat[:,3]]  # (kg m-2)
    return state_inst


def get_jules_state(nc_file):
    """Function that returns a stateVector instance for a given time.

    :param date_utc: datetime object of when to extract JULES output.
    :type date_utc: object
    :param nc_file: JULES output file from which to extract data.
    :type nc_file: str
    :return: Instance of stateVector class.
    :rtype: instance
    """
    nc_dat = nc.Dataset(nc_file, 'r')
    state_inst = StateVector()
    state_inst.date_utc = nc.num2date(nc_dat.variables['time'][:], nc_dat.variables['time'].units).tolist()
    state_inst.lai = nc_dat.variables['croplai'][:, 0, 0, 0].tolist()  # (m2 m-2)
    state_inst.can_height = nc_dat.variables['cropcanht'][:, 0, 0, 0].tolist()  # (m)
    state_inst.soil_moisture = (nc_dat.variables['smcl'][:, 0, 0, 0]/100).tolist() # (kg m-2)
    nc_dat.close()
    return state_inst


def get_jules_state_old(date_utc, nc_file):
    """Function that returns a stateVector instance for a given time.

    :param date_utc: datetime object of when to extract JULES output.
    :type date_utc: object
    :param nc_file: JULES output file from which to extract data.
    :type nc_file: str
    :return: Instance of stateVector class.
    :rtype: instance
    """
    nc_dat = nc.Dataset(nc_file, 'r')
    t_idx = nc.date2index(date_utc, nc_dat.variables['time'], select='nearest')
    state_inst = StateVector()
    state_inst.date_utc = nc.num2date(nc_dat.variables['time'][t_idx], nc_dat.variables['time'].units)
    state_inst.lai = nc_dat.variables['croplai'][t_idx, 0, 0, 0]  # (m2 m-2)
    state_inst.can_height = nc_dat.variables['cropcanht'][t_idx, 0, 0, 0]  # (m)
    state_inst.soil_moisture = nc_dat.variables['smcl'][t_idx, 0, 0, 0]  # (kg m-2)
    nc_dat.close()
    return state_inst


def find_nearest_date_idx(items, pivot):
    nearest = min(items, key=lambda x: abs(x - pivot))
    timedelta = abs(nearest - pivot)
    if timedelta.days > 31:
        warnings.warn('Your closest land state in time is specified over a month before your specified satellite ' \
                      'observation')
    idx = np.where(np.array(items)==nearest)[0][0]
    return idx, timedelta


def get_date_list(year, month=1, days=365):
    start_date = dt.datetime(year, month, 1, 12, 0)
    date_list = pd.date_range(start_date, periods=days).tolist()
    return date_list


def read(file_format='jules', file_str=None, year=None):
    """Reads output data to a dictionary of state vectors indexed by time.

    .. note:: This function requires sub-functions capable of reading specified file format.

    :param file_format: format of output to read.
    :type file_str: str
    :param file_str: location of file.
    :type file_str: str
    :param year: year of data to extract, if equal to None whole time series extracted
    :type year: int
    :return: state dictionary.
    :rtype: dict
    """
    if file_format == 'jules':
        state_dict = read_jules(file_str, year)
    else:
        state_dict = {}
    return state_dict


def read_jules(nc_file=None, year=None):
    """Reads jules output from netCDF file and writes it to a dictionary indexed by date.

    :param nc_file: location of nc_file.
    :type nc_file: str
    :param year: year of data to extract, if equal to None whole time series extracted.
    :type year: int
    :return: state dictionary.
    :rtype: dict
    """
    if nc_file is None:
        nc_file = 'jules/output/wallerfing_jules_1989_2012.nc'
        print("%s") % nc_file
    nc_dat = nc.Dataset(nc_file, 'r')
    state_dict = {}
    time = nc_dat.variables['time']
    if year is not None:
        strt_idx = nc.date2index(dt.datetime(year, 1, 1), time)
        end_idx = nc.date2index(dt.datetime(year, 12, 31), time)
        t_idx = np.arange(strt_idx, end_idx + 1)
        times = nc.num2date(time[strt_idx:end_idx], time.units)
    else:
        times = nc.num2date(time[:], time.units)
        t_idx = np.arange(len(times))
    for t in enumerate(times):
        state_dict[t[1]] = StateVector()
        state_dict[t[1]].lai = nc_dat.variables['croplai'][t_idx[t[0]], 0, 0, 0]  # (m2 m-2)
        state_dict[t[1]].can_height = nc_dat.variables['cropcanht'][t_idx[t[0]], 0, 0, 0]  # (m)
        state_dict[t[1]].soil_moisture = nc_dat.variables['smcl'][t_idx[t[0]], 0, 0, 0]  # (kg m-2)
        # state_dict[t[1]].soil_temp = nc_dat.variables['t_soil'][t_idx[t[0]], 0, 0, 0]  # (K)
        # figure out how to add others, which soil albedo to output?
    return state_dict



# python builtins:
import os
import sys
import subprocess
from copy import copy
from tempfile import mkstemp

# third party imports:
import netCDF4
import numpy as np

# sentienl synergy specific imports:
import spectra as sp
import satellite_geometry as satgeo
import state_vector as sv


def canopy_rt_optical(state, geom, resln=1.0):
    """A python wrapper to the SemiDiscrete optical
    canopy RT model of Nadine Gobron. Runs the
    model for the the whole of its valid spectra
    range at a resolution set by resln.

    :param state: Instance of the stateVector class.
    :type state: instance
    :param geom: Instance of the sensorGeomety class.
    :type geom: instance
    :param resln: the spectral resolution in nm [optional].
    :type resln: float
    :return: Instance of the spectra class.
    :rtype: instance
    """

    spect = sp.spectra()
    spect.wavl = np.arange(400, 2500 + resln, resln)

    # generate a tmp file and write
    # wavelengths and geometry into it
    fd, temp_path = mkstemp(prefix='/tmp/senSyntmp__', text=True)
    tmpFile = os.fdopen(fd, 'w')
    print >> tmpFile, "1 %d" % len(spect.wavl),
    for w in spect.wavl:
        print >> tmpFile, " %f" % w,
    print >> tmpFile, "\n%s %s %s %s" %(geom.vza,geom.vaa,geom.sza,geom.saa)
    tmpFile.close()

    # set up nadim command line
    cmd = "semiDiscrete"
    if state.lai != None:
        cmd = cmd + " -LAI %f -hc %f -rsl1 %f" %(state.lai, state.can_height, 0.2*(1.-0.5*(state.soil_moisture/100.)))
        # Think about soil moisture implementation here
    cmd = cmd + " < %s" % temp_path

    # run process
    p = subprocess.Popen(cmd, stdout=subprocess.PIPE, stderr=sys.stderr, shell=True)
    out = p.stdout.readlines()
    p.wait()

    # clean up
    os.remove(temp_path)

    # process RT model output
    rtOut = out[1].split()
    for i in xrange(4, len(rtOut)):
        reflTmp = np.append(spect.refl, float(rtOut[i]))
        spect.refl = copy(reflTmp)

    return spect


def canopy_rt_optical_fast(state, geom, resln=1.0, mode='fast'):
    """A python wrapper to the SemiDiscrete optical
    canopy RT model of Nadine Gobron. Runs the
    model for the the whole of its valid spectra
    range at a resolution set by resln.

    :param state: Instance of the StateVector class.
    :type state: instance
    :param geom: Instance of the SensorGeometry class.
    :type geom: instance
    :param mode: Run semiDiscrete in either fast ('fast') or slow ('slow') mode [optional].
    :type resln: str
    :return: Instance of the spectra class.
    :rtype: instance
    """
    spect = sp.spectra()
    spect.wavl = np.arange(400, 2500 + resln, resln)

    # generate a tmp file and write geometry into it
    fd, temp_path = mkstemp(prefix='/tmp/senSyntmp__', text=True)
    tmpFile = os.fdopen(fd, 'w')
    print >> tmpFile, "%s %s %s %s" %(geom.vza,geom.vaa,geom.sza,geom.saa)
    # print "%s %s %s %s" %(geom.vza,geom.vaa,geom.sza,geom.saa)
    tmpFile.close()

    # set up nadim command line
    if mode == 'fast':
        cmd = "semiD -srf ../data/srf/s2a.srf -fast"
    else:
        cmd = "semiD -srf ../data/srf/s2a.srf"
    if state.lai != None:
        cmd = cmd + " -LAI %f -hc %f -rsl1 %f" %(state.lai, state.can_height, 0.2*(1.-0.5*(state.soil_moisture/100.)))
        # Think about soil moisture implementation here
    cmd = cmd + " < %s" % temp_path
    # print cmd

    # run process
    p = subprocess.Popen(cmd, stdout=subprocess.PIPE, stderr=sys.stderr, shell=True)
    out = p.stdout.readlines()
    p.wait()
    spect.refl = np.array([float(val) for val in out[0].split()])
    return spect


def passive_optical_rt(state, geom):
    """Function that simulates reflectances given surface biogeophysical variables and viewing geometries.

    :param state: Instance of the StateVector class.
    :type state: instance
    :param geom: Instance of the SensorGeometry class.
    :type geom: instance
    :return: Instance of the spectra class.
    :rtype: instance
    """
    spect = sp.spectra()
    spect.date_sat_ob = []
    spect.date_land_ob = []
    spect.soil_moisture = []
    spect.lai = []
    spect.can_height = []
    spect.refl = []

    for date_utc in enumerate(geom.date_utc):
        idx, timedelt = sv.find_nearest_date_idx(state.date_utc, date_utc[1])
        reflectance = run_semidiscrete(state.soil_moisture[idx], state.lai[idx], state.can_height[idx],
                                       geom.vza[date_utc[0]], geom.vaa[date_utc[0]], geom.sza[date_utc[0]],
                                       geom.saa[date_utc[0]])
        spect.date_sat_ob.append(date_utc[1])
        spect.date_land_ob.append(state.date_utc[idx])
        spect.soil_moisture.append(state.soil_moisture[idx])
        spect.lai.append(state.lai[idx])
        spect.can_height.append(state.can_height[idx])
        spect.refl.append(reflectance)
    return spect


def run_semidiscrete(soil_m, lai, can_height, vza, vaa, sza, saa, mode='fast'):
    """A python wrapper to the SemiDiscrete optical canopy RT model of Nadine Gobron.

    :param soil_m: soil moisture at specified time (m3 m-3)
    :type soil_m: float
    :param lai: Leaf area index at specified time (m2 m-2)
    :type lai: float
    :param can_height: canopy height at specified time (m)
    :type can_height: float
    :param vza: view zenith angle (degrees)
    :type vza: float
    :param vaa: view azimuth angle (degrees)
    :type vaa: float
    :param sza: solar zenith angle (degrees)
    :type sza: float
    :param saa: solar azimuth angle (degrees)
    :type saa: float
    :param mode: Run semiDiscrete in either fast ('fast') or slow ('slow') mode [optional].
    :type resln: str
    :return: Reflectance values for specified input.
    :rtype: array
    """

    # generate a tmp file and write geometry into it
    dir_path = os.path.dirname(os.path.realpath(__file__))
    fd, temp_path = mkstemp(prefix='/tmp/senSyntmp__', text=True)
    tmpFile = os.fdopen(fd, 'w')
    print >> tmpFile, "%s %s %s %s" % (vza, vaa, sza, saa)
    # print "%s %s %s %s" %(geom.vza,geom.vaa,geom.sza,geom.saa)
    tmpFile.close()

    # set up nadim command line
    if mode == 'fast':
        cmd = dir_path+"/semidiscrete_srf/semiD -srf "+dir_path+"/data/srf/s2a.srf -fast"
    else:
        cmd = dir_path+"/semidiscrete_srf/semiD -srf "+dir_path+"/data/srf/s2a.srf"
    if lai != None:
        cmd = cmd + " -LAI %f -hc %f -rsl1 %f" % (lai, can_height, 0.2 * (1. - 0.5 * soil_m))
        # Think about soil moisture implementation here
    cmd = cmd + " < %s" % temp_path
    # print cmd

    # run process
    p = subprocess.Popen(cmd, stdout=subprocess.PIPE, stderr=sys.stderr, shell=True)
    out = p.stdout.readlines()
    p.wait()

    reflectance = np.array([float(val) for val in out[0].split()])
    return reflectance


if __name__ == "__main__":
    """This example opens a test output file from the JULES
    model, reads in LAI data, runs these through the optical RT
    model at given geometry and convoles the resulting spectra
    with Sentinel 2 spectra response functions.
    """

    from matplotlib import pyplot as plt

    # read example LAI data:
    allLai = netCDF4.Dataset('../testData/crop_germ_gl4.day.nc').variables['lai'][-365:, 5, 0, 0]

    # main classes:
    state = sv.stateVector()
    geom = satgeo.sensorGeometry()
    geom.sza = 30.

    # container for output:
    allSpect = []
    allBRF = []

    # orbit revist:
    revist = 10

    # plotting variables:
    xpnts = []
    lgnd = []

    # loop over all states and call the
    # canopy RT model and convolve the
    # retruned spectra to S2 bands
    for (n, L) in enumerate(allLai[::revist]):
        state.lai = L
        spect = canopy_rt_optical(state, geom)
        allSpect.append(sp.sentinel2(spect))
        allBRF.append(allSpect[n].refl)
        xpnts.append(n * revist)

    allBRF = np.array(allBRF)

    # sort out legend
    for i in xrange(len(allBRF[0, :])):
        lgnd.append('S2 band %d' % (i + 1))

    # do plots:
    lineObjs = plt.plot(xpnts, allBRF)
    plt.ylabel('reflectance (BRF)')
    plt.xlabel('Day of year')
    plt.xlim([0, 364])
    for i in xrange(7, len(lineObjs)):
        lineObjs[i].set_dashes([3, 1])
    plt.legend(iter(lineObjs), lgnd)
    plt.show()
    # plt.savefig('s2Sim_test1.png')



"""
Dielectric mixing model for soils
after Dobson et al. (1985)
coding after Ulaby (2014), Chapter 4
"""

import numpy as np
from . epsmodel import EpsModel

class Dobson85(EpsModel):
    def __init__(self, **kwargs):
        super(Dobson85, self).__init__(**kwargs)

        self._init_model_parameters()
        self.ew = self._calc_ew()
        self.eps = self._calc_eps()

    def _calc_ew(self, debye=False):
        """
        calculate dielectric permittivity of free water
        using either the Debye model or a more simplistic approach
        """
        if debye:
            return self._debye()
        else:
            # simplistic approach using Eq. 4.69
            return self._simple_ew()

    def _simple_ew(self):
        """ eq. 4.69 """
        f0 = 18.64   # relaxation frequency [GHz]
        hlp = self.f/f0
        e1 = 4.9 + (74.1)/(1.+hlp**2.)
        e2 =(74.1*hlp)/(1.+hlp**2.) + 6.46 * self.sigma/self.f 
        return e1 + 1.j * e2

    def _debye(self):
        assert False


    def _init_model_parameters(self):
        """
        model parameters, eq. 4.68, Ulaby (2014)
        """
        self.alpha = 0.65
        self.beta1 = 1.27-0.519*self.sand - 0.152*self.clay
        self.beta2 = 2.06 - 0.928*self.sand -0.255*self.clay
        self.sigma = -1.645 + 1.939*self.bulk - 2.256*self.sand + 1.594*self.clay

    def _calc_eps(self):
        """
        calculate dielectric permittivity
        Eq. 4.66 (Ulaby et al., 2014)
        """

        e1 = (1.+0.66*self.bulk+self.mv**self.beta1*np.real(self.ew)**self.alpha - self.mv)**(1./self.alpha)
        e2 = np.imag(self.ew)*self.mv**self.beta2
        return e1 + 1.j*e2
        



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# Filename: __init__.py
# Author: "Thomas Weiß"
# Date:   2017-07-26 13:24:17
# Last Modified by:   "Thomas Weiß"
# Last Modified time: 2017-07-26 13:26:08

"""
Information about the content
"""






"""
Class specifying a soil
"""
import numpy as np
from . util import f2lam
from . dielectric import Dobson85

class Soil(object):
    def __init__(self, **kwargs):
        """
        Parameters
        ----------
        eps : complex
            relative permitivity, if this is not given, then mv needs to be given
        s : float
            surface rms height [m]
        mv : float
            volumetric soil moisture [m**3/m**3]; either eps or mv needs to be given
        f : float
            frequency [GHz]
        l : float
            optional: autocorrelation length
        acl : str
            identifier for sphape of autocorrelation fucntion 
            G = Gaussian
            E = Exponential
        mv : float
            volumetric moisture content [m**3/m**3]
        clay : float
            optional fractional clay content
        sand : float
            optional fraction sand content
        """

        self.eps = kwargs.get('eps', None)
        self.mv = kwargs.get('mv', None)
        self.f = kwargs.get('f', None)
        self.s = kwargs.get('s', None)
        self.l = kwargs.get('l', None)
        self.acl = kwargs.get('acl', None)
        self.clay = kwargs.get('clay', None)
        self.sand = kwargs.get('sand', None)
        self.dc_model = kwargs.get('dc_model', 'Dobson85')
        self._check()
        if self.eps is not None:
            self._convert_eps2mv()
        if self.mv is not None:
            self._convert_mv2eps()

        # wavenumber
        self.k = 2.*np.pi / f2lam(self.f)  # note that wavenumber is in meter and NOT in cm!
        
        # roughness parameters
        self.ks = self.s*self.k
        if self.l is not None:
            self.kl = self.k*self.l
        else:
            self.kl = None


    def _convert_mv2eps(self):
        """
        convert mv to eps 
        using dielectric model
        """
        if (self.clay is None) or (self.sand is None):
            self.eps = None
            print('WARNING: Permittivity can not be calculated due to missing soil texture!')
        if self.dc_model == 'Dobson85':
            DC = Dobson85(clay=self.clay, sand=self.sand, mv=self.mv, freq=self.f) 
        else:
            assert False, 'Invalid DC model! ' + self.dc_model
        
        self.eps = DC.eps



    def _convert_eps2mv(self):
        """
        This routine converts soil moisture into
        dielectric properties and vice versa

        future implementations will comprise e.g. the Dobson model
        and others ...


        """
        assert self.eps is not None, 'Currently cnversion not implemented yet; you need to provide the DC directly!'

    def _check(self):
        if self.acl is not None:
            assert self.acl in ['G','E'], 'Invalid form of autocorrelation function specified'
        assert self.s is not None

        if self.eps is None:
            assert self.mv is not None, 'Either EPS or MV need to be given!'
        if self.mv is None:
            assert self.eps is not None, 'Either EPS or MV need to be given!'







#!/usr/bin/env python
# -*- coding: utf-8 -*-
# Filename: scatter.py
# Author: "Thomas Weiß"
# Date:   2017-07-26 13:46:39
# Last Modified by:   "Thomas Weiß"
# Last Modified time: 2017-07-26 13:48:55

"""
Major vegetation scatter class
"""

class VegetationScatterWaterCloudModel(object):
    def __init__(self, theta, V=None, V1=None, V2=None, A=None, B=None):
        self.theta = theta
        self.V = V
        self.V1 = V1
        self.V2 = V2
        self.A = A
        self.B = B




"""
Generic model for dielectric mixing models
"""

import numpy as np

class EpsModel(object):
    def __init__(self, **kwargs):
        """
        Parameters
        ----------
        clay : float
            clay content as fractional volume
        sand : float
            sand content as fractional volume
        bulk : float
            bulk density [g/cm**3]; default: 1.65
        mv : float
            volumetric soil moisture content [cm**3/cm**3] = [g/cm**3]
        freq : float
            frequency [GHz]
        """

        self.clay = kwargs.get('clay', None)
        self.sand = kwargs.get('sand', None)

        self.bulk = kwargs.get('bulk', 1.65)

        self.mv = kwargs.get('mv', None)

        self.f = kwargs.get('freq', None)

        self._check()

    def _check(self):
        assert self.clay is not None, 'Clay needs to be provided!'
        assert self.clay >=0.
        assert self.clay <=1.
        assert self.sand is not None, 'Sand needs to be provided!'
        assert self.sand >=0.
        assert self.sand <=1.
        assert self.mv is not None, 'volumetric soil moisture needs to be given'
        assert self.f is not None, 'Frequency needs to be given!'
        if isinstance(self.f, np.ndarray):
            assert np.all(self.f> 0.)
        else:
            assert self.f > 0.



"""
This module implements the Oh et al. (1992)
empirical surface backscattering model as documented
in Ulaby et al (2014), chapter 10.5

References
----------
Oh et al. (1992): An empirical model and an inversion technique for radar scattering from bare soil surfaces. IEEE TGRS 30(2). 370-381.
"""

import numpy as np
import matplotlib.pyplot as plt

from . scatter import SurfaceScatter
from .. core import Fresnel0
from .. core import Reflectivity

class Oh92(SurfaceScatter):
    def __init__(self, eps, ks, theta):
        """
        Parameters
        ----------
        eps : complex
            relative dielectric permitivity
        ks : float
            product of wavenumber and rms height
            be aware that both need to have the same units
        theta : float, ndarray
            incidence angle [rad]
        """
        super(Oh92, self).__init__(eps, ks, theta)

        # calculate p and q
        self.G0 = Fresnel0(self.eps)  # nadir fresnel reflectivity
        self.G = Reflectivity(self.eps, self.theta)
        self._calc_p()
        self._calc_q()

        # calculate backscatter
        # (added 10*np.log10() like in PRISM1_FORWARDMODEL-1.m)
        self._vv0 = self._calc_vv()
        self.vv = self._vv0
        self.hh = self.p * self._vv0
        self.hv = self.q * self._vv0

    def _calc_p(self):
        a = 1./(3.*self.G0.x)
        self.p = (1. - (2.*self.theta/np.pi)**a * np.exp(-self.ks))**2.

    def _calc_q(self):
        self.q = 0.23*(self.G0.x)**0.5 * (1.-np.exp(-self.ks))

    def _calc_vv(self):

        a = 0.7*(1.-np.exp(-0.65*self.ks**1.8))
        b = np.cos(self.theta)**3. * (self.G.v+self.G.h) / np.sqrt(self.p)
        return a*b

    def plot(self):
        f = plt.figure()
        ax = f.add_subplot(111)
        t = np.rad2deg(self.theta)
        ax.plot(t, 10*np.log10(self.hh), color='blue', label='hh')
        ax.plot(t, 10*np.log10(self.vv), color='red', label='vv')
        ax.plot(t, 10*np.log10(self.hv), color='green', label='hv')
        ax.grid()
        ax.set_ylim(-25.,0.)
        ax.set_xlim(0.,70.)
        ax.legend()
        ax.set_xlabel('incidence angle [deg]')
        ax.set_ylabel('backscatter [dB]')










def fwd_atm(spectra):
    """
    Given an instance of the spectra class function will forward model reflectances to top of atmosphere
    :param spectra: class instance of spectra
    :type spectra: object
    :return: updated spectra instance
    """
    return spectra


# third party imports
import numpy as np
# signature simulator specific imports
import state_vector as sv
import backscatter as bs
# import sense code
from sense import model as sense_mod
from sense import soil as sense_soil
from sense import canopy as sense_canopy


def active_microwave_rt(state, geom):
    """
    Function that simulates SAR backscattering, given surface biogeophysical variables and viewing geometries
    :param state: instance of StateVector class
    :type state: object
    :param geom: instance of SensorGeometry class
    :type geom: object
    :return: instance of BackScatter class
    """
    backscat = bs.BackScatter()
    backscat.date_sat_ob = []
    backscat.date_land_ob = []
    backscat.soil_moisture = []
    backscat.lai = []
    backscat.can_height = []
    backscat.hh = []
    backscat.hv = []
    backscat.vv = []
    for date_utc in enumerate(geom.date_utc):
        idx, timedelt = sv.find_nearest_date_idx(state.date_utc, date_utc[1])
        SAR = run_sense(state.soil_moisture[idx], state.lai[idx], state.can_height[idx], geom.vza[date_utc[0]])
        backscat.date_sat_ob.append(date_utc[1])
        backscat.date_land_ob.append(state.date_utc[idx])
        backscat.soil_moisture.append(state.soil_moisture[idx])
        backscat.lai.append(state.lai[idx])
        backscat.can_height.append(state.can_height[idx])
        backscat.hh.append(10 * np.log10(SAR.__dict__['stot']['hh']))
        backscat.hv.append(10 * np.log10(SAR.__dict__['stot']['hv']))
        backscat.vv.append(10 * np.log10(SAR.__dict__['stot']['vv']))
    return backscat



def run_sense(soil_m, lai, can_height, vza, freq=5.405, stype='turbid_rayleigh', surf='Oh92', s=0.015, lai_coef=0.1,
                        omega=0.1):
    """
    Function that runs the SenSE SAR ScattEring model given some inputs
    :param soil_m: soil moisture (m3 m-3)
    :type soil_m: float
    :param lai: leaf area index (m2 m-2)
    :type lai: float
    :param can_height: canopy height (m)
    :type can_height: float
    :param vza: view zenith angle of sensor (degrees)
    :type vza: float
    :param freq: frequency (GHz)
    :type freq: float
    :param stype: canopy scattering model
    :type stype: str
    :param surf: surface scattering model
    :type surf: str
    :param s: surface rms height (m)
    :type s: float
    :param lai_coef: coefficient of lai for which to calculate extinction and volume scattering coefficients
    :type lai_coef: float
    :param omega: coefficient for calculation of volume scattering coefficients
    :type omega: float
    :return: instance of sense SingleScatRT class
    """
    """Function that runs sense model, given surface biogeophysical variables and viewing geometries
    """
    models = {'surface': surf, 'canopy': stype}
    SAR = sense_mod.SingleScatRT(
          surface=sense_soil.Soil(mv=soil_m, f=freq, s=s, clay=0.23, sand=0.27),
          # surface=sense_soil.Soil(eps=self.eps, f=self.freq, s=self.s),
          canopy=sense_canopy.OneLayer(ke_h=lai_coef * lai,  # extinction coefficient
                                       ke_v=lai_coef * lai,
                                       d=can_height,
                                       ks_h=omega * (lai_coef * lai),  # volume scattering coeff
                                       ks_v=omega * (lai_coef * lai)),
          models=models,
          theta=np.deg2rad(vza),
          freq=freq)
    SAR.sigma0()
    return SAR


import numpy as np
import matplotlib.pyplot as plt
import matplotlib.dates as mdates
import core as sim


def plot_class_var(date_lst, var, y_lab=None, line_type='-', axes=None):
    """Plot specified variable.

    :param var: Class attribute variable as list.
    :type var: list
    :param y_lab: Label for Y-axis.
    :type y_lab: str
    :return: Figure.
    :rtype: object
    """
    if axes is not None:
        ax = axes
        ret_val = ax
    else:
        fig, ax = plt.subplots(nrows=1, ncols=1, figsize=(15, 5))
        ret_val = fig, ax
    #ax.xaxis_date()
    ax.plot(date_lst, var, line_type)
    plt.ylabel(y_lab)
    plt.xlabel('Date')
    plt.gcf().autofmt_xdate()
    myFmt = mdates.DateFormatter('%B')
    ax.xaxis.set_major_formatter(myFmt)
    # plt.legend(loc=2)
    # plt.show()
    return ret_val


def plot_backscat(lai_coeff, s, omega):
    """Plot backscatter for given coefficients.

    :param lai_coeff: LAI constant used in calculation of extinction coefficient
    :type lai_coeff: float
    :param s: Surface roughness
    :type s: float
    :param omega: Single scattering albedo
    :type omega: float
    :return: Figure
    :rtype: object
    """
    """
    """
    sim_c1 = sim.S1_simulator(lai_coef=lai_coeff, s=s, omega=omega, jules_nc='jules/output/demo/test.3_hourly.nc')
    for x in xrange(3):
        fig = plot_class_var(sim_c1.date_lst,
                             [10*np.log10(sim_c1.SAR_list[s1c].__dict__['stot'][sim_c1.backscatter_keys[x]])
                              for s1c in xrange(len(sim_c1.SAR_list))],
                             y_lab='Backscatter ' + sim_c1.backscatter_keys[x] + ' polarisation (db)',
                             line_type='o')[0]
        # Must also think about canopy height and extinction coefficient!!!
        fig.savefig('jules/output/demo/epstest_' + sim_c1.backscatter_keys[x] + '_laicoeff' + str(np.round(lai_coeff,4))
                    + '.png')
        plt.close('all')
    plt.close('all')
    return 'done'


"""
implements the I2EM model (see Ulaby (2014), Chapter 10
backscattering model for single scale random surfaces


The code originates from ideas obtained from the supplement
of Ulaby et al (2014)
"""
from . scatter import SurfaceScatter
import numpy as np

from .. util import f2lam
from .. core import Reflectivity
import math

from scipy.integrate import dblquad


# from numba import jit


#@jit(cache=True,nopython=True)
def _calc_roughness_spectra_matrix(nx, ny, kl2, nspec, s, acf_type_id):
    """
    calculate roughness spectra
    needs to return a matrix for further use
    in crosspol calculations
    """

    if acf_type_id == 1:  # gauss
        wm = _calc_wm_matrix_gauss(nx, ny, nspec, kl2, s)
        wn = _calc_wn_matrix_gauss(nx, ny, nspec, kl2, s)

    elif acf_type_id == 2:  # exp
        wm = _calc_wm_matrix_exp(nx, ny, nspec, kl2, s)
        wn = _calc_wn_matrix_exp(nx, ny, nspec, kl2, s)
    else:
        assert False
    return wn, wm





class I2EM(SurfaceScatter):
    def __init__(self, f, eps, sig, l, theta, **kwargs):
        """

        BACKSCATTERING MODEL

        Parameters
        ----------
        f : float
            frequency [GHz]
        eps : complex
            relative dielectric permitivity
        sig : float
            vertical surface roughness  [m]
        l : float
            autocorrelation length [m]
        theta : float
            incidence angle [rad]
        acf_type : str
            type of autocorrelation function
            'gauss' : gaussian type ACF
        auto : bool
            specify if number of spectral components should be automatically
            determined for cross-pol calculations
            if False, then nspec=15
        xpol : bool
            perform cross-pol calculations if possible
            might be slow in case of I2EM usage
        """

        self.freq = f
        lam = f2lam(self.freq)
        k = 2.*np.pi/lam
        self.k = k
        self.sig = sig
        self.ks = self.k*self.sig
        self.l = l
        self._kl2 = (self.k*self.l)**2.
        self.acf_type = kwargs.get('acf_type', 'gauss')
        super(I2EM, self).__init__(eps, k*sig, theta, kl=k*l)
        
        # assume backscatter geometry
        self.phi = 0.
        self.thetas = self.theta*1.
        self.phis = np.deg2rad(180.)
        self.mode = 'backscatter'

        self.auto = kwargs.get('auto', True)
        self.xpol = kwargs.get('xpol', True)

        # do initializations for backscatter calculations
        self._init_hlp()
        self.init_model()

        # calculate the actual backscattering coefficients
        self._calc_sigma_backscatter()

    def init_model(self):
        """
        initialize model for calculations
        """
        self.niter = self._estimate_itterations()

        # determine number of spectral components for cross-pol calculations
        if self.auto:
            # same as function _estimate_itterations, but with slightly different config
            nspec = 0
            error = 1.E8
            while error > 1.0E-8:
                nspec += 1
                error = (self._ks2*(2.*self._cs)**2.)**nspec / math.factorial(nspec)  
            self.n_spec = nspec
        else:
            self.n_spec = 15

        I = np.arange(self.n_spec)
        self._fac = map(math.factorial, I+1)  # factorial(n)



    def _estimate_itterations(self):
        """
        estimate the number of necessary itterations for 
        the integral calculations
        """

        err = 1.E8
        Ts = 1
        while err > 1.0e-8:
            Ts += 1
            err = ((self._ks2 *(self._cs + self._css)**2 )**Ts) / math.factorial(Ts)
        return Ts


    def _init_hlp(self):
        """ initiate help variables """
        self._ks2 = self.ks**2.
        self._cs = np.cos(self.theta)
        self._cs2 = self._cs**2.
        self._s = np.sin(self.theta)
        self._sf = np.sin(self.phi)
        self._cf = np.cos(self.phi)
        self._ss = np.sin(self.thetas)
        self._css = np.cos(self.thetas)
        self._cfs = np.cos(self.phis)
        self._sfs = np.sin(self.phis)
        self._s2 = self._s**2.
        self._kx = self.k*self._s*self._cf
        self._ky = self.k*self._s*self._sf
        self._kz = self.k*self._cs

        self._ksx = self.k * self._ss *self._cfs
        self._ksy = self.k * self._ss *self._sfs
        self._ksz = self.k * self._css

    def _calc_sigma_backscatter(self):
        assert isinstance(self.theta, float), 'Currently array processing not supported yet!'
        # calculate backscattering coefficients
        self.vv, self.hh = self._i2em_bistatic()
        if self.xpol:
            self.hv = self._i2em_cross()

    def _i2em_bistatic(self):
        """
        calculate sigma for the co-pol case
        backscatter geometr
        calculate sigma for the co-pol case
        backscatter geometry

        module 10.1
        """

        # calculate the integral
        idx = np.arange(self.niter)+1
        self.fac = map(math.factorial, idx)  # factorial for all N itterations; this is stored as it is needed multipole times

        self.wn, self.rss = self.calc_roughness_spectrum(acf_type=self.acf_type) 
        Ivv, Ihh = self._calc_Ipp()
        Ivv_abs = np.abs(Ivv)
        Ihh_abs = np.abs(Ihh)

        # calculate shadowing effects
        ShdwS = self._calc_shadowing()

        a0 = self.wn / self.fac * (self.sig**(2.*idx))

        # final backscatter calculation
        hlp = ShdwS*0.5*self.k**2*np.exp(-self.sig**2*(self._kz**2.+self._ksz**2.))
        sigvv = np.sum(a0 * Ivv_abs**2.) * hlp
        sighh = np.sum(a0 * Ihh_abs**2.) * hlp
        return  sigvv, sighh

    def _i2em_cross(self):
        rt = np.sqrt(self.eps - self._s2)
        rv = (self.eps*self._cs -rt) / (self.eps*self._cs + rt)
        rh = (self._cs - rt)/(self._cs + rt)
        rvh = (rv-rh)/2.

        Shdw = self._calc_shadow_cross()

        svh = self._integrate_xpol(rvh)
        return svh*Shdw


    def _integrate_xpol(self, rvh):
        """
        integrate for X-pol
        dblquad(@(r,phi)xpol_integralfunc(r, phi, sp,xx, ks2, cs,s, kl2, L, er, rss, rvh, n_spec), 0.1, 1, 0, pi)

        the original matlab routines integrates
        xpol_integral(r,phi)
        rmin=0.1, rmax=1.
        phimin=0.,phimax=1.

        when using python, x and y are reversed, however
        this does not matter unless the bounds are specified in the right order
        """
        ans, err = dblquad(self._xpol_integralfunc, 0.1, 1., lambda x : 0., lambda x : 1., args=[[rvh,self.eps, self._ks2, self._cs2, self.rss, self._cs, self._fac, self._kl2, self._s, self._get_acf_id()]])
        return ans



    def _get_acf_id(self):
        if self.acf_type == 'gauss':
            return 1
        if self.acf_type == 'exp15':
            return 2
        assert False, 'Unknown ACF type'


    #@jit(cache=True)
    def _xpol_integralfunc(self, r, phi, *args):
        """
        while the original matlab function
        returns a vector, this function
        returns a scalar, as the dblquad function
        in python requires so
        """

        rvh = args[0][0]
        eps = args[0][1]
        ks2 = args[0][2]
        cs2 = args[0][3]
        rss = args[0][4]
        cs = args[0][5]
        fac = args[0][6]
        nspec = len(fac)
        kl2 = args[0][7]
        s = args[0][8]
        acf_type_id = args[0][9]

        r2 = r**2.
        sf = np.sin(phi)
        csf = np.cos(phi)
        rx = r * csf
        ry = r * sf

        rp = 1. + rvh
        rm = 1. - rvh

        q = np.sqrt(1.0001 - r2)
        qt = np.sqrt(eps - r2)

        a = rp / q
        b = rm / q
        c = rp / qt
        d = rm / qt

        # calculate cross-pol coefficient
        B3 = rx * ry / cs
        fvh1 = (b-c)*(1.- 3.*rvh) - (b - c/eps) * rp
        fvh2 = (a-d)*(1.+ 3.*rvh) - (a - d*eps) * rm
        Fvh = ( np.abs( (fvh1 + fvh2) *B3))**2.

        # calculate x-pol shadowing
        au = q /r /1.414 /rss
        fsh = (0.2821/au) *np.exp(-au**2.) -0.5 *(1.- math.erf(au))
        sha = 1./(1. + fsh)

        # calculate expressions for the surface spectra
        
        wn, wm = _calc_roughness_spectra_matrix(rx, ry, kl2, nspec, s, acf_type_id) 

        vhmnsum = 0.
        for i in xrange(nspec):
            for j in xrange(nspec):
                vhmnsum += wn[i] * wm[j] * (ks2*cs2)**((i+1)+(j+1))/fac[i]/fac[j] 

        # compute VH scattering coefficient
        acc = np.exp(-2.* ks2 *cs2) /(16. * np.pi)
        VH = 4. * acc * Fvh * vhmnsum * r
        y = VH * sha
        return y




    def _calc_shadow_cross(self):
        """"
        calculating shadow consideration in single scat (Smith, 1967)
        """
        ct = np.cos(self.theta)/np.sin(self.theta)
        farg = ct /np.sqrt(2.) /self.rss
        gamma = 0.5 *(np.exp(-farg**2.) / 1.772 / farg - math.erfc(farg))
        return 1. / (1. + gamma)

    def _calc_shadowing(self):

        if self.mode == 'backscatter':    #todo comparison with binary variable instead of string to be faster ??

            ct = np.cos(self.theta)/np.sin(self.theta)
            cts = np.cos(self.thetas)/np.sin(self.thetas)
            rslp = self.rss
            ctorslp = ct / math.sqrt(2.) /rslp
            ctsorslp = cts / np.sqrt(2.) /rslp
            shadf = 0.5 *(np.exp(-ctorslp**2.) / np.sqrt(np.pi)/ctorslp - math.erfc(ctorslp))
            shadfs = 0.5 *(np.exp(-ctsorslp**2.) / np.sqrt(np.pi)/ctsorslp - math.erfc(ctsorslp))
            ShdwS = 1./(1. + shadf + shadfs)
        else:
            ShdwS = 1.

        return ShdwS

    def calc_roughness_spectrum(self, acf_type=None):
        """
        calculate roughness spectrum
        Return wn as an array
        """
        assert 'Validate with code again'
        if acf_type == 'gauss':
            # gaussian autocorrelation function
            S = GaussianSpectrum(niter=self.niter, l=self.l, theta=self.theta, thetas=self.thetas, phi=self.phi,phis=self.phis, freq=self.freq, sig=self.sig)
        elif acf_type == 'exp15':
            # 1.5 power exponential function
            S = ExponentialSpectrum(niter=self.niter, l=self.l, theta=self.theta, thetas=self.thetas, phi=self.phi,phis=self.phis, freq=self.freq, sig=self.sig)
        else:
            assert False, 'Invalid surface roughness spectrum: ' + str(acf_type)
        
        return S.wn()  # returns wn as an array with length NITER

    def _calc_Ipp(self):
        n = np.arange(self.niter)+1.
        qi = self.k*self._cs
        qs = self.k*self._css

        h1= np.exp(-self.sig**2. * self._kz * self._ksz)*(self._kz + self._ksz)**n

        # Calculate the Fppup(dn) i(s) coefficient
        R = Reflectivity(self.eps, self.theta)
        Rvi = R.rho_v
        Rhi = R.rho_h
        
        Fvvupi, Fhhupi = self.Fppupdn( 1,1,Rvi,Rhi)
        Fvvups, Fhhups = self.Fppupdn( 1,2,Rvi,Rhi)
        Fvvdni, Fhhdni = self.Fppupdn(-1,1,Rvi,Rhi)
        Fvvdns, Fhhdns = self.Fppupdn(-1,2,Rvi,Rhi)

        # fpp calculations
        fvv, fhh = self.calc_fpp(Rvi, Rhi)

        # Ipp
        Ivv = fvv*h1
        Ivv += 0.25*(Fvvupi *(self._ksz-qi)**(n-1) *np.exp(-self.sig**2. *(qi**2. - qi*(self._ksz-self._kz))))
        Ivv += 0.25*(Fvvdni *(self._ksz+qi)**(n-1) *np.exp(-self.sig**2. *(qi**2. + qi*(self._ksz-self._kz))))
        Ivv += 0.25*(Fvvups *(self._kz +qs)**(n-1) *np.exp(-self.sig**2. *(qs**2. - qs*(self._ksz-self._kz))))
        Ivv += 0.25*(Fvvdns *(self._kz -qs)**(n-1) *np.exp(-self.sig**2. *(qs**2. + qs*(self._ksz-self._kz))))

        Ihh = fhh*h1
        Ihh += 0.25*(Fhhupi *(self._ksz-qi)**(n-1) *np.exp(-self.sig**2. *(qi**2. - qi*(self._ksz-self._kz))))
        Ihh += 0.25*(Fhhdni *(self._ksz+qi)**(n-1) *np.exp(-self.sig**2. *(qi**2. + qi*(self._ksz-self._kz))))
        Ihh += 0.25*(Fhhups *(self._kz +qs)**(n-1) *np.exp(-self.sig**2. *(qs**2. - qs*(self._ksz-self._kz))))
        Ihh += 0.25*(Fhhdns *(self._kz -qs)**(n-1) *np.exp(-self.sig**2. *(qs**2. + qs*(self._ksz-self._kz))))

        return Ivv, Ihh

    def calc_fpp(self, Rvi, Rhi):

        Rvt, Rht = self.calc_reflection_coefficients(Rvi, Rhi)

        fvv =  2. * Rvt *(self._s * self._ss - (1. + self._cs * self._css) * self._cfs)/(self._cs + self._css)
        fhh = -2. * Rht *(self._s * self._ss - (1. + self._cs * self._css) * self._cfs)/(self._cs + self._css)
        return fvv, fhh


    def Fppupdn(self, u_d, i_s, Rvi, Rhi):
        assert i_s in [1,2]
        assert u_d in [-1,1]

        # set coefficients
        if i_s == 1:
            Gqi = u_d * self._kz
            Gqti = u_d * self.k *np.sqrt(self.eps-self._s**2.);
            qi = u_d * self._kz
            c11 = self.k * self._cfs *(self._ksz - qi)
            c21 = self._cs *(self._cfs *(self.k**2 *self._s*self._cf*(self._ss *self._cfs - self._s * self._cf) + Gqi*(self.k * self._css - qi))+ self.k**2. *self._cf * self._s *self._ss *self._sfs**2.)

            c31 = self.k*self._s*(self._s*self._cf*self._cfs*(self.k*self._css-qi) - Gqi*(self._cfs*(self._ss*self._cfs -self._s*self._cf)+ self._ss *self._sfs**2.))
            c41 = self.k *self._cs*(self._cfs*self._css*(self.k*self._css - qi) + self.k *self._ss*(self._ss*self._cfs-self._s*self._cf))
            c51 = Gqi*(self._cfs *self._css*(qi-self.k*self._css) - self.k *self._ss*(self._ss*self._cfs-self._s*self._cf))
            c12 = self.k * self._cfs *(self._ksz - qi)

            c22 = self._cs *(self._cfs *(self.k**2. *self._s*self._cf*(self._ss *self._cfs - self._s * self._cf) + Gqti*(self.k * self._css - qi)) + self.k**2. *self._cf * self._s *self._ss *self._sfs**2.)
            c32 = self.k*self._s*(self._s*self._cf*self._cfs*(self.k*self._css-qi) - Gqti*(self._cfs*(self._ss*self._cfs -self._s*self._cf)- self._ss *self._sfs**2.))

            c42 = self.k *self._cs*(self._cfs*self._css*(self.k*self._css - qi) + self.k *self._ss*(self._ss*self._cfs-self._s*self._cf))

            c52 = Gqti*(self._cfs *self._css*(qi-self.k*self._css) - self.k *self._ss*(self._ss*self._cfs-self._s*self._cf))

        else:
            Gqs = u_d * self._ksz
            Gqts = u_d *self.k *np.sqrt(self.eps-self._ss**2.)
            qs = u_d * self._ksz

            c11 = self.k * self._cfs *(self._kz + qs)
            c21 = Gqs *(self._cfs*(self._cs*(self.k*self._cs+qs)-self.k*self._s*(self._ss *self._cfs-self._s*self._cf))-self.k*self._s*self._ss*self._sfs**2.)
            c31 = self.k *self._ss*(self.k*self._cs*(self._ss*self._cfs - self._s*self._cf)+ self._s*(self._kz+qs))
            c41 = self.k*self._css*(self._cfs*(self._cs*(self._kz+qs)-self.k*self._s*(self._ss*self._cfs-self._s*self._cf))-self.k*self._s*self._ss*self._sfs**2.)
            c51 = -self._css *(self.k**2. *self._ss *(self._ss*self._cfs -self._s*self._cf)+ Gqs*self._cfs*(self._kz+qs))
            c12 = self.k * self._cfs *(self._kz + qs)
            c22 = Gqts *(self._cfs*(self._cs*(self._kz+qs)-self.k*self._s*(self._ss *self._cfs-self._s*self._cf))-self.k*self._s*self._ss*self._sfs**2.)
            c32 = self.k *self._ss*(self.k*self._cs*(self._ss*self._cfs - self._s*self._cf)+ self._s*(self._kz+qs))
            c42 = self.k*self._css*(self._cfs*(self._cs*(self._kz+qs)-self.k*self._s*(self._ss*self._cfs-self._s*self._cf))-self.k*self._s*self._ss*self._sfs**2.)
            c52 = -self._css *(self.k**2. *self._ss *(self._ss*self._cfs -self._s*self._cf)+ Gqts*self._cfs*(self._kz+qs))


        # now do final calculations ...
        q = self._kz
        qt = self.k * np.sqrt(self.eps - self._s**2.)

        vv =  (1.+Rvi) *( -(1-Rvi) *c11 /q + (1.+Rvi)       *c12 / qt)
        vv += (1.-Rvi) *(  (1-Rvi) *c21 /q - (1.+Rvi)       *c22 / qt)
        vv += (1.+Rvi) *(  (1-Rvi) *c31 /q - (1.+Rvi)       *c32 /self.eps /qt) 
        vv += (1.-Rvi) *(  (1+Rvi) *c41 /q - self.eps*(1. - Rvi)  *c42 / qt)
        vv += (1.+Rvi) *(  (1+Rvi) *c51 /q - (1.-Rvi)       *c52 / qt)

        hh =  (1.+Rhi) *( (1.-Rhi) * c11 /q - self.eps *(1.+Rhi) *c12 / qt)
        hh -= (1.-Rhi) *( (1.-Rhi) * c21 /q - (1.+Rhi)    *c22 / qt)
        hh -= (1.+Rhi) *( (1.-Rhi) * c31 /q - (1.+Rhi)    *c32 / qt)
        hh -= (1.-Rhi) *( (1.+Rhi) * c41 /q - (1.-Rhi)    *c42 / qt)
        hh -= (1.+Rhi) *( (1.+Rhi) * c51 /q - (1.-Rhi)    *c52 / qt)

        return vv, hh


    def _calc_r_transition(self):
        """ compute R transition """

        Rv0 = (np.sqrt(self.eps)-1.) / (np.sqrt(self.eps) + 1.)
        Rh0 = -Rv0

        Ft = 8. * Rv0**2. + self._ss * (self._cs + np.sqrt(self.eps - self._s2))/(self._cs * np.sqrt(self.eps - self._s2))

        idx = np.arange(self.niter)+1
        a0 = (self.ks*self._cs)**(2.*idx)/self.fac
        a1 = np.sum(a0*self.wn)
        b1 = np.sum(a0 * (np.abs(Ft/2. + 2.**(idx+1) *Rv0/self._cs *np.exp(-(self.ks*self._cs)**2.)))**2. * self.wn)

        St = 0.25 * np.abs(Ft)**2. * a1/b1
        St0 = 1. / np.abs(1.+8.*Rv0/(self._cs * Ft))**2.
        Tf = 1. - St / St0

        return Rv0, Rh0, Tf

    def _calculate_average_reflection_coefficients(self):
        assert False, 'Not implemented yet!'
        #%----------- compute average reflection coefficients ------------
        #%-- these coefficients account for slope effects, especially near the
        #%brewster angle. They are not important if the slope is small.

        #sigx = 1.1 .*sig/L;
        #sigy = sigx;
        #xxx = 3*sigx;

        #Rav = dblquad(@(Zx, Zy)Rav_integration(Zx, Zy, cs,s,er,s2,sigx, sigy),-xxx,xxx, -xxx, xxx );

        #Rah = dblquad(@(Zx, Zy)Rah_integration(Zx, Zy, cs,s,er,s2,sigx, sigy),-xxx,xxx, -xxx, xxx );

        #Rav = Rav ./(2*pi * sigx * sigy);
        #Rah = Rah ./(2*pi * sigx * sigy);



    def calc_reflection_coefficients(self, Rvi, Rhi):


        Rv0, Rh0, Tf = self._calc_r_transition()


        # select proper reflection coefficients
        if self.mode == 'backscatter':  # todo this comparison might slow down the program as it is called very often; perhaps modify
            Rvt = Rvi + (Rv0 - Rvi) * Tf
            Rht = Rhi + (Rh0 - Rhi) * Tf
        elif self.mode == 'bistatic':
            Rav = Rah = self._calculate_average_reflection_coefficients()
            Rvt = Rav
            Rht = Rah
            pass
        else:
            assert False

        return Rvt, Rht


class Roughness(object):
    """
    calculate roughness spectrum
    """
    def __init__(self, **kwargs):
        self.niter = kwargs.get('niter', None)
        self.l = kwargs.get('l', None)
        self.sig = kwargs.get('sig', None)
        self.theta = kwargs.get('theta', None)
        self.thetas = kwargs.get('thetas', None)
        self.phi = kwargs.get('phi', None)
        self.phis = kwargs.get('phis', None)
        self.freq = kwargs.get('freq', None)
        
        self._check()
        self.n = np.arange(self.niter)+1
        self._init()

    def wn(self):
        assert False, 'Should be implemented in child class!'

    def _init(self):
        ss = np.sin(self.thetas)
        self._s = np.sin(self.theta)
        sf = np.sin(self.phi)
        sfs = np.sin(self.phis)
        cfs = np.cos(self.phis)
        cf = np.cos(self.phi)
        lam = f2lam(self.freq)
        self.k = 2.*np.pi / lam
        self._kl = self.k*self.l
        self._kl2 = self._kl**2.
        
        # todo whereis this defined ???
        self.wvnb = self.k * np.sqrt( (ss *cfs - self._s *cf)**2. + (ss * sfs - self._s * sf)**2. )

    def _check(self):
        assert self.niter is not None, 'ERROR: niter was not set!'
        assert self.l is not None
        assert self.sig is not None 
        assert self.theta is not None
        assert self.thetas is not None
        assert self.phi is not None
        assert self.phis is not None
        assert self.freq is not None



#@jit(cache=False, nopython=True)
def _calc_wn_matrix_gauss(rx, ry, nspec, kl2, s):
    wn = np.zeros(nspec)
    for i in xrange(nspec):
        wn[i] = 0.5 *kl2/(i+1.) * np.exp(-kl2*((rx-s)**2. + ry**2.)/(4.*(i+1))) 
    return wn

#@jit(cache=False, nopython=True)
def _calc_wm_matrix_gauss(rx, ry, nspec, kl2, s):
    wm = np.zeros(nspec)
    for i in xrange(nspec):
        wm[i] = 0.5 *kl2/(i+1.) * np.exp(-kl2*((rx+s)**2. + ry**2.)/(4.*(i+1))) 
    return wm

class GaussianSpectrum(Roughness):
    def __init__(self, **kwargs):
        super(GaussianSpectrum, self).__init__(**kwargs)
    
    def wn(self):
        # Fung (1994), Eq. 2B.4; except for wvnb
        n = self.n
        wn = (self.l**2.)/(2.*n) * np.exp(-(self.wvnb*self.l)**2. / (4.*n))
        rss = np.sqrt(2.)*self.sig/self.l
        return wn, rss

    def calc_wn_matrix(self, rx, ry, nspec):
        return _calc_wn_matrix_gauss(rx, ry, nspec, self._kl2, self._s)

    def calc_wm_matrix(self, rx, ry, nspec):

        return _calc_wm_matrix_gauss(rx, ry, nspec, self._kl2, self._s)

#@jit(cache=True,nopython=True)
def _calc_wn_matrix_exp(rx, ry, nspec, kl2, s):
    wn = np.zeros(nspec)
    for i in xrange(nspec):
        wn[i] = (i+1) * kl2 / ((i+1)**2.+kl2*((rx-s)**2. + ry**2.))**1.5
    return wn

#@jit(cache=True,nopython=True)
def _calc_wm_matrix_exp(rx, ry, nspec, kl2, s):
    wm = np.zeros(nspec)
    for i in xrange(nspec):
        wm[i] = (i+1) * kl2 / ((i+1)**2.+kl2*((rx+s)**2. + ry**2.))**1.5
    return wm


class ExponentialSpectrum(Roughness):
    """
    exponential spectrum
    """
    def __init__(self, **kwargs):
        super(ExponentialSpectrum, self).__init__(**kwargs)

    def wn(self):
        # Fung (1994): eq. 2.B.14
        n = self.n
        wn= self.l**2. / n**2. * (1.+(self.wvnb*self.l/n)**2.)**(-1.5)
        rss = self.sig/self.l
        return wn, rss
    
    def calc_wn_matrix(self, rx, ry, nspec):
        #for i in xrange(nspec):
        # n = i+1
        #return np.array([(i+1) * self._kl2 / ((i+1)**2.+self._kl2*((rx-self._s)**2. + ry**2.))**1.5 for i in xrange(nspec)])
        return _calc_wn_matrix_gauss(rx, ry, nspec, self._kl2, self._s)

    def calc_wm_matrix(self, rx, ry, nspec):
        #return np.array([(i+1) * self._kl2 / ((i+1)**2.+self._kl2*((rx+self._s)**2. + ry**2.))**1.5 for i in xrange(nspec)])
        return _calc_wm_matrix_gauss(rx, ry, nspec, self._kl2, self._s)






import numpy as np
import matplotlib.pyplot as plt

class Fresnel0(object):
    def __init__(self, e):
        """
        calculate the Nadir Fresnel reflectivity
        e.g. Ulaby (2014), eq. 10.36

        Parameters
        ----------
        e : complex
            complex relative dielectric permitivity
        """
        self.x = self._calc(e)

    def _calc(self, e):
        return np.abs( (1.-np.sqrt(e))/(1.+np.sqrt(e))   )**2.



class Reflectivity(object):
    """
    calculate the reflectivity for H and V polarization
    """
    def __init__(self, eps, theta):
        """
        table 2.5 Ulaby (2014)
        assumes specular surface
        Parameters
        ----------
        eps : complex
            relative dielectric permitivity
        theta : float, ndarray
            incidence angle [rad]
            can be specified
        """
        self.eps = eps
        self.theta = theta

        self._calc_reflection_coefficients()

        self.v = np.abs(self.rho_v)**2.
        self.h = np.abs(self.rho_h)**2.


    def _calc_reflection_coefficients(self):
        """
        calculate reflection coefficients
        Woodhouse, 2006; Eq. 5.54, 5.55
        """
        # OLD
        co = np.cos(self.theta)
        si2 = np.sin(self.theta)**2.
        self.rho_v = (self.eps*co-np.sqrt(self.eps-si2))/(self.eps*co+np.sqrt(self.eps-si2))
        self.rho_h = (co-np.sqrt(self.eps-si2))/(co+np.sqrt(self.eps-si2))

#        srv = self.rho_v
#        srh = self.rho_h
#
#        # FROM PRISM1_FORWARDMODEL-1.m
#        n1 = np.sqrt(1.)
#        n2 = np.sqrt(self.eps)
#        costh2 = np.sqrt(1-(n1*np.sin(self.theta)/2.)**2)
#
#        self.rho_v = -(n2*np.cos(self.theta) - n1*costh2)/(n2*np.cos(self.theta) + n1*costh2)
#        self.rho_h = (n1*np.cos(self.theta) - n2*costh2)/(n1*np.cos(self.theta) + n2*costh2)

#        plt.plot(np.rad2deg(self.theta), self.rho_v-srv, label = 'v_diff')
#        plt.plot(np.rad2deg(self.theta), self.rho_h-srh, label = 'h_diff')
#        plt.legend()
        #doesn't make much difference in results!

    def plot(self):
        f = plt.figure()
        ax = f.add_subplot(111)
        ax.plot(np.rad2deg(self.theta), self.v, color='red', linestyle='-', label='V')
        ax.plot(np.rad2deg(self.theta), self.h, color='blue', linestyle='--', label='H')
        ax.grid()
        ax.legend()
















"""
implements the Dubois95 model
as described in Ulaby (2014), Chapter 10.6
"""
import numpy as np
import matplotlib.pyplot as plt

from . scatter import SurfaceScatter

class Dubois95(SurfaceScatter):
    def __init__(self, eps, ks, theta, lam=None):
        """
        Parameters
        ----------
        lam : float
            wavelength in meter
        """

        super(Dubois95, self).__init__(eps, ks, theta)
        self.lam = lam
        assert self.lam is not None
        self.vv, self.hh = self._calc_sigma()
        self.hv = None

    def _calc_sigma(self):
        lam = self.lam*100.  # from m in cm
        ks = self.ks/100  # This line reproduces ULABY value. s is in cm, here!
        return self._vv(lam, ks), self._hh(lam, ks)

    def _hh(self, lam, ks):
        """
        lam : float
            wavelength in cm
        """

        a = (10.**-2.75)*(np.cos(self.theta)**1.5)/(np.sin(self.theta)**5.)
        c = 10.**(0.028*np.real(self.eps)*np.tan(self.theta))
        d = ((ks*np.sin(self.theta))**1.4)*lam**0.7

        return a*c*d

    def _vv(self, lam, ks):
        """ eq. 10.41b """
        b = 10.**(-2.35)*((np.cos(self.theta)**3.) / (np.sin(self.theta)**3.))
        c = 10.**(0.046*np.real(self.eps)*np.tan(self.theta))
        d = (ks*np.sin(self.theta))**1.1*lam**0.7

        return b*c*d

    def plot(self):
        f = plt.figure()
        ax = f.add_subplot(111)
        t = np.rad2deg(self.theta)
        ax.plot(t, 10.*np.log10(self.hh), color='blue', label='hh')
        ax.plot(t, 10.*np.log10(self.vv), color='red', label='vv')
       # ax.plot(t, 10.*np.log10(self.hv), color='green', label='hv')
        ax.grid()
        ax.set_ylim(-35.,-5.)
        ax.set_xlim(30.,70.)
        ax.legend()
        ax.set_xlabel('incidence angle [deg]')
        ax.set_ylabel('backscatter [dB]')



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# Filename: prevot1993.py
# Author: "Thomas Weiß"
# Date:   2017-07-26 13:25:10
# Last Modified by:   "Thomas Weiß"
# Last Modified time: 2017-07-26 13:55:33

"""
This module implements the Prevot et al. (1993) empirical vegetation backscattering model component

References
----------
Prevot et al. (1992): Estimation the characteristics of vegetation canopy with airborne radar measurements
"""

import numpy as np
import matplotlib.pyplot as plt

from . scatter import VegetationScatterWaterCloudModel

class Prevot93_vegetation(VegetationScatterWaterCloudModel):
    def __init__(self, V, A, B, theta):
        """
        Parameters
        ----------
        V : float
            water content of the canopy [kg/m**2]
        A : float
            correspond to the albedo of the vegetation
        B : float
            attenuation factor
        theta : float
            incidence angle [rad]
        """
        super(Prevot93_vegetation, self).__init__(theta, V=V, A=A, B=B)

        # calculate backscatter
        self.tau = np.exp(-2 * self.B * self.V / np.cos(np.radians(self.theta)))
        self.vegetation = self.A * np.cos(np.radians(self.theta)) * (1 - self.tau)

    def plot(self):
        f = plt.figure()
        ax = f.add_subplot(111)
        t = np.rad2deg(self.theta)
        ax.plot(t, 10.*np.log10(self.vegetation), color='blue', label='vegetation component')
        ax.grid()
        # ax.set_ylim(-25.,0.)
        # ax.set_xlim(0.,70.)
        ax.legend()
        ax.set_xlabel('incidence angle [deg]')
        ax.set_ylabel('backscatter [dB]')
        plt.show()
        pdb.set_trace()







"""
Definition of scatter types
"""
import numpy as np

class Scatterer(object):
    def __init__(self, **kwargs):
        # NOTE THAT THE arguments are not necessarily the
        # particle scattering cross sections!!!!
        # need to be made in a clearer way !!!!
        self.sigma_s_hh = kwargs.get('sigma_s_hh', None)  # particle scattering cross area
        assert self.sigma_s_hh is not None, 'Particle HH scattering cross section needs to be specified [m**-2]'

        self.sigma_s_vv = kwargs.get('sigma_s_vv', None)  # particle scattering cross area
        assert self.sigma_s_vv is not None, 'Particle VV scattering cross section needs to be specified [m**-2]'

        self.sigma_s_hv = kwargs.get('sigma_s_hv', None)  # particle scattering cross area
        assert self.sigma_s_hv is not None, 'Particle HV scattering cross section needs to be specified [m**-2]'


class ScatIso(Scatterer):
    """
    Isotropic scatterer definition
    see 11.2 in Ulaby (2014)
    """
    def __init__(self, **kwargs):
        super(ScatIso, self).__init__(**kwargs)
        # scattering coefficients need to be the same in isotropic case (eq. 11.19)
        assert self.sigma_s_hh == self.sigma_s_vv
        assert self.sigma_s_hh == self.sigma_s_hv

    def sigma_v_back(self):
        """
        volume backscattering coefficient
        for the isotropic case this corresponds to the
        volume scattering coefficient ks

        not that this is NOT the scattering cross section of a single particle!
        """
        return {'hh' : self.sigma_s_hh, 'vv' : self.sigma_s_vv, 'hv' : self.sigma_s_hv}

    def sigma_v_bist(self):
        # same as volume backscattering coefficient (Eq. 11.19)
        return self.sigma_v_back()



class ScatRayleigh(Scatterer):
    """
    Isotropic scatterer definition
    see 11.2 in Ulaby (2014)
    """
    def __init__(self, **kwargs):
        super(ScatRayleigh, self).__init__(**kwargs)

    def sigma_v_back(self):
        # sigma_s_pp is assumed to correspond to volume extinction coefficient
        return {'hh' : 1.5*self.sigma_s_hh, 'vv' : 1.5*self.sigma_s_vv, 'hv' : np.nan}

    def sigma_v_bist(self):
        # same as sigma_v_back (Eq. 11.22)
        return self.sigma_v_back()



"""
Basic class for scattering modelling
"""

import numpy as np
from . surface import Oh92, Dubois95
from . util import f2lam
from . scatterer import ScatIso, ScatRayleigh
from . core import Reflectivity

class Model(object):
    def __init__(self, **kwargs):
        self.theta = kwargs.get('theta', None)

        self._check1()

    def _check1(self):
        assert self.theta is not None, 'ERROR: no incidence angle was specified!'

    def sigma0(self, **kwargs):
        """
        calculate sigma

        Parameters
        ----------
        dB : bool
            return results in decibel
        pol : list
            list with polarizations pq
            whereas p=receive, q=transmit
            p,g can be either H or V
        """

        self.dB = kwargs.get('dB', False)
        self.pol = kwargs.get('pol', [])
        self._check_pol()

        if self.dB:
            assert False, 'Not supported for dictionaries yet!'
            return 10.*np.log10(self._sigma0())
        else:
            return self._sigma0()

    def _sigma0(self, **kwargs):
        assert False, 'routine should be implemented in child class!'

    def _check_pol(self):
        if len(self.pol) == 0:
            assert 'ERROR: polarization needs to be specified'
        for k in self.pol:
            if k not in ['HH','VV','HV','VH']:
                assert False, 'Invalid polarization: ' + k



class SingleScatRT(Model):
    def __init__(self, **kwargs):
        """
        Single scattering model according to Ulaby and Long (2014)
        Eq. 11.17

        Parameters
        ----------
        surface : Surface description
            object describing the surface
        canopy : Canopy description
            object describing the canopy
        models : dict
            dictionary with configuration of scattering models
        """
        super(SingleScatRT, self).__init__(**kwargs)
        self.surface = kwargs.get('surface', None)
        self.canopy = kwargs.get('canopy', None)
        self.models = kwargs.get('models', None)
        self.freq = kwargs.get('freq', None)
        self.coherent = kwargs.get('coherent', True)  # use coherent simulations as default

        self._check()

    def _check(self):
        assert self.surface is not None
        assert self.canopy is not None
        assert self.models is not None
        assert self.freq is not None

        for k in ['surface', 'canopy']:
            assert k in self.models.keys()  # check that all models have been specified

        assert self.freq == self.surface.f, "Different frequencies in model and soil definition"
            # check that frequencies are the same!

    def _sigma0(self):
        """
        basic calculation of Sigma0
        based on Eq. 11.17 in Ulaby and Long (2014)
        """

        # ground backscatter = attenuated surface
        self.G = Ground(self.surface, self.canopy, self.models['surface'], self.models['canopy'], theta=self.theta, freq=self.freq)
        self.s0g = self.G.sigma()  # returns dictionary with different components

        # canopy contribution
        self.s0c = self.G.rt_c.sigma_c()   # returns a dictionary

        # total canopy ground contribution
        self.s0cgt = self.G.sigma_c_g(self.coherent)

        # ground-canopy-ground interaction
        self.s0gcg = self.G.sigma_g_c_g()

        # combine backscatter values
        self.stot = {}
        for k in ['hh', 'vv', 'hv']:
            self.stot.update({k : self._combine(k)})

    def _combine(self, k):
        """
        combine previous calculated backscatter values
        """
        if self.s0g[k] is None:
            return None
        if self.s0c[k] is None:
            return None
        return np.nansum(np.array([self.s0g[k], self.s0c[k], self.s0gcg[k], self.s0cgt[k]]))

class Ground(object):
    """
    calculate the (attenuated) ground contribution
    sigma_pq
    where p is receive and q is transmit polarization
    """
    def __init__(self, S, C, RT_s, RT_c, theta=None, freq=None):
        """
        calculate the attenuated ground contribution
        to the scattering

        Parameters
        ----------
        S : object
            descibing the surface properties
        C : object
            describing the canopy properties
        RT_s : str
            key describing the surface scattering model
        RT_c : str
            key specifying the canopy scattering model
        freq : float
            frequency[GHz]
        """
        self.S = S
        self.C = C
        self.theta = theta
        self._check(RT_s, RT_c)
        self.freq = freq
        assert self.freq is not None, 'Frequency needs to be provided'
        self._set_models(RT_s, RT_c)
        self._calc_rho()

    def _calc_rho(self):
        """
        calculate coherent p-polarized
        reflectivity
        Ref: Eq. 11.11 (Ulaby, 2014)

        Note that the specular reflectivity is corrected by a roughness term
        if ks>0.2

        however, a sensitivity analysis showed that even for ks==0.2
        deviations can be up to 15% for typical incidence angles
        Only in case that ks << 0.1, the correction can be neglected.
        We therefore always use the roughness correction factor!

        TODO: unclear so far how this relates to surface (soil) scattering models
        """

        R = Reflectivity(self.S.eps, self.theta)
        self.rho_v = R.v * np.exp(-4.*np.cos(self.theta)**2.*(self.S.ks**2.))
        self.rho_h = R.h * np.exp(-4.*np.cos(self.theta)**2.*(self.S.ks**2.))

    def _set_models(self, RT_s, RT_c):
        # set surface model
        if RT_s == 'Oh92':
            self.rt_s = Oh92(self.S.eps, self.S.ks, self.theta)
        elif RT_s == 'Dubois95':
            self.rt_s = Dubois95(self.S.eps, self.S.ks, self.theta, lam=f2lam(self.freq))
        else:
            assert False, 'Unknown surface scattering model'


        # set canopy models
        if RT_c == 'turbid_isotropic':  # turbid media (homogenous vegetation)
            self.rt_c = CanopyHomoRT(ke_h=self.C.ke_h, ke_v=self.C.ke_v, ks_h=self.C.ks_h, ks_v=self.C.ks_v, d=self.C.d, theta=self.theta, stype='iso')
        elif RT_c == 'turbid_rayleigh':
            self.rt_c = CanopyHomoRT(ke_h=self.C.ke_h, ke_v=self.C.ke_v, ks_h=self.C.ks_h, ks_v=self.C.ks_v, d=self.C.d, theta=self.theta, stype='rayleigh')
        else:
            assert False, 'Invalid canopy scattering model: ' + RT_c

    def _check(self, RT_s, RT_c):
        valid_surface = ['Oh92', 'Dubois95']
        valid_canopy = ['turbid_rayleigh', 'turbid_isotropic']
        assert RT_s in valid_surface, 'ERROR: invalid surface scattering model was chosen!'
        assert RT_c in valid_canopy, 'ERROR: invalid canopy model: ' + RT_c
        assert self.theta is not None

    def sigma_g_c_g(self):


        s_vv = self.rt_c.sigma_vol_back['vv']*np.cos(self.theta)*self.rho_v*self.rho_v*(self.rt_c.t_v*self.rt_c.t_v-self.rt_c.t_v**4.) / (self.C.ke_v + self.C.ke_v)
        s_hh = self.rt_c.sigma_vol_back['hh']*np.cos(self.theta)*self.rho_h*self.rho_h*(self.rt_c.t_h*self.rt_c.t_h-self.rt_c.t_h**4.) / (self.C.ke_h + self.C.ke_h)
        s_hv = self.rt_c.sigma_vol_back['hv']*np.cos(self.theta)*self.rho_h*self.rho_v*(self.rt_c.t_h*self.rt_c.t_v-self.rt_c.t_h**2.*self.rt_c.t_v**2.) / (self.C.ke_h + self.C.ke_v)

        return {'vv' : s_vv, 'hh' : s_hh, 'hv' : s_hv}


    def sigma_c_g(self, coherent=None):
        """
        calculate canopy ground scattering coefficient
        This is based on Eq. 11.17 (last term) in Ulaby (2014)
        and 11.14 in Ulaby (2014)

        for co-pol, coherent addition can be made as an option

        Parameters
        ----------
        coherent : bool
            do coherent calculation for co-pol calculations
        """
        assert coherent is not None, 'ERROR: please specify explicity if coherent calculations should be made.'
        if coherent:
            n = 2.
        else:
            n = 1.

        s_vv = n  * self.rt_c.sigma_vol_bistatic['vv'] * self.C.d *(self.rho_v + self.rho_v)*self.rt_c.t_v*self.rt_c.t_v
        s_hh = n  * self.rt_c.sigma_vol_bistatic['hh'] * self.C.d *(self.rho_h + self.rho_h)*self.rt_c.t_h*self.rt_c.t_h
        s_hv = 1. * self.rt_c.sigma_vol_bistatic['hv'] * self.C.d *(self.rho_v + self.rho_h)*self.rt_c.t_h*self.rt_c.t_v


        return {'vv' : s_vv, 'hh' : s_hh, 'hv' : s_hv}



    def sigma(self):
        """
        calculate the backscattering coefficient
        Eq. 11.4, p.463 Ulaby (2014)
        """

        # canopy transmisivities
        t_h = self.rt_c.t_h
        t_v = self.rt_c.t_v

        # backscatter
        s_hh = self.rt_s.hh*t_h*t_h
        s_vv = self.rt_s.vv*t_v*t_v

        if self.rt_s.hv is None:
            s_hv = None
        else:
            s_hv = self.rt_s.hv*t_v*t_h


        return {'vv' : s_vv, 'hh' : s_hh, 'hv' : s_hv}


class CanopyHomoRT(object):
    """
    homogeneous canopy RT model
    assumes homogeneous vertical distribution of scatterers

    in that case the Lambert Beer law applies

    NOTE that this model is only for BACKSCATTERING GEOMETRY!
    """
    def __init__(self, **kwargs):
        """
        Parameters
        ----------
        ke_h, ke_v : float
            volume extinction coefficient [Np/m]
        d : float
            height of canopy layer
        theta : float, ndarray
            incidence angle [rad]
        """
        self.ke_h = kwargs.get('ke_h', None)
        self.ke_v = kwargs.get('ke_v', None)
        self.ks_h = kwargs.get('ks_h', None)
        self.ks_v = kwargs.get('ks_v', None)
        self.theta = kwargs.get('theta', None)
        self.d = kwargs.get('d', None)
        #self.Nv = kwargs.get('Nv', 1.)
        self.stype = kwargs.get('stype', None)  # scatterer type

        self._check()

        self.tau_h = self._tau(self.ke_h)
        self.tau_v = self._tau(self.ke_v)
        self.t_h = np.exp(-self.tau_h)
        self.t_v = np.exp(-self.tau_v)

        self._set_scat_type()
        self.sigma_vol_back = self._calc_back_volume()
        self.sigma_vol_bistatic = self._calc_sigma_bistatic()

    def _check(self):
        assert self.stype is not None

        assert self.ke_h is not None
        assert self.ke_v is not None
        assert self.ks_h is not None
        assert self.ks_v is not None

        assert self.ke_h >=0.
        assert self.ke_v >=0.
        assert self.ks_h >=0.
        assert self.ks_v >=0.


        assert self.ks_h <= self.ke_h
        assert self.ks_v <= self.ke_v

    def _set_scat_type(self):
        """ set scatterer type """
        if self.stype == 'iso':
            self.SC = ScatIso(sigma_s_hh=self.ks_h, sigma_s_vv=self.ks_v, sigma_s_hv=self.ks_v)   # note that the cross pol scatt. coeff. is the same as the copol due to isotropic behavior
        elif self.stype == 'rayleigh':
            self.SC = ScatRayleigh(sigma_s_hh = self.ks_h, sigma_s_vv=self.ks_v, sigma_s_hv=self.ks_v)  # eq. 11.22
        elif self.stype == 'cloud':
            assert False  # here implemenatation of 11.5 then
        else:
            assert False, 'Invalid scatterer type specified: ' + self.stype

    def _calc_back_volume(self):
        """
        calculate the volume backscattering coefficient sigma_v
        This is a function of the scatterer type chosen (e.g. isotropic,
        rayleigh, cloud model, ...)
        """
        return self.SC.sigma_v_back()

    def _calc_sigma_bistatic(self):
        """
        calculate volume bistatic scattering coefficient
        of scatterer
        """
        return self.SC.sigma_v_bist()




    def _tau(self, k):
        # assumption: extinction is isotropic
        return k*self.d/np.cos(self.theta)

    def sigma_gcg(self, G_v, G_h):
        """
        calculate ground-canopy-ground interactions
        Eq. 11.16, Ulaby(2014)

        Parameters
        ----------
        G_v : float
            v-polarized coherent Fresnel reflectivity under rough conditions
            see eq. 11.11 for explanations. As this depends on the
            surface model used, these should be provided here explicitely
        G_h : float
            same as above, but for h-polarization.
        """
        return G_v*G_h*(self.t_h*self.t_v-self.t_h**2.*self.t_v**2.)*(self.sigma_vol*np.cos(self.theta))/(self.ke_h+self.ke_v)

    def sigma_c(self):
        """
        calculate canopy volume contribution only
        Eq. 11.10 + 11.16 as seen in 11.17, Ulaby (2014)
        """

        s_hh = (1.-self.t_h*self.t_h)*(self.sigma_vol_back['hh']*np.cos(self.theta))/(self.ke_h+self.ke_h)
        s_vv = (1.-self.t_v*self.t_v)*(self.sigma_vol_back['vv']*np.cos(self.theta))/(self.ke_v+self.ke_v)
        s_hv = (1.-self.t_h*self.t_v)*(self.sigma_vol_back['hv']*np.cos(self.theta))/(self.ke_h+self.ke_v)

        # this seems o.k. here
#        a=self.sigma_vol_back['hh']
#        b=1.5*self.ks_h
#        print a,b,a-b, a/b, self.ks_h

        return {'hh' : s_hh, 'vv' : s_vv, 'hv' : s_hv}


# 502-503






from pyorbital.orbital import Orbital
import datetime as dt
import pyorbital.astronomy as ast
import numpy as np


class SensorGeometry:
    """Class to hold sun-sensor geometry information.
    """
    def __init__(self):
        self.date_utc = None
        # view zenith angle
        self.vza = None
        # view azimuth angle
        self.vaa = None
        # solar zenith angle
        self.sza = None
        # solar azimuth angle
        self.saa = None

    def print_geom(self):
        """Prints currently specified class attributes.
        """
        print(self.date_utc, self.vza, self.vaa, self.sza, self.saa)


def get_geom_csv(fname='/data/geometries/s2_example.csv'):
    """Function that returns StateVector instance for a given file
    :param fname: Path to file from which to extract data, files must be csv with columns:
     (date, lai, canopy height, soil moisture)
    :type fname: str
    :return: StateVector instance
    """
    geom_dat = np.loadtxt(fname, delimiter=',', dtype='str')
    geom_inst = SensorGeometry()
    geom_inst.date_utc = [dt.datetime.strptime(dd, "%Y/%m/%d %H:%M") for dd in geom_dat[:,0]]
    geom_inst.vza = [float(vza) for vza in geom_dat[:,1]]
    geom_inst.vaa = [float(vaa) for vaa in geom_dat[:,2]]
    geom_inst.sza = [float(sza) for sza in geom_dat[:,3]]
    geom_inst.saa = [float(saa) for saa in geom_dat[:,4]]
    return geom_inst


def get_satellite_geometry(startDateUTC, lengthDays, lon, lat, alt=0.0, mission="Sentinel-2a",
                           tleFile="/data/tle/norad_resource_tle.txt"):
    """Calculate approximate geometry for Sentinel overpasses.
    Approximate because it assumes maximum satellite elevation
    is the time at which target is imaged.

    :param startDateUTC: a datetime object specifying when to start prediction.
    :type startDateUTC: object
    :param lengthDays: number of days over which to perform calculations.
    :type lengthDays: int
    :param lat: latitude of target.
    :type lat: float
    :param lon: longitude of target.
    :type lon: float
    :param alt: altitude of target (in km).
    :type alt: float
    :param mission: mission name as in TLE file.
    :type mission: str
    :param tleFile: TLE file.
    :type tleFile: str
    :return: a python list containing instances of the sensorGeometry class arranged in date order.
    :rtype: list
    """
    orb = Orbital(mission, tleFile)
    passes = orb.get_next_passes(startDateUTC, 24 * lengthDays, lon, lat, alt)

    geom_inst = SensorGeometry()
    geom_inst.date_utc = []
    geom_inst.vza = []
    geom_inst.vaa = []
    geom_inst.sza = []
    geom_inst.saa = []

    for p in passes:
        look = orb.get_observer_look(p[2], lon, lat, alt)
        vza = 90 - look[1]
        vaa = look[0]
        sza = ast.sun_zenith_angle(p[2], lon, lat)
        saa = np.rad2deg(ast.get_alt_az(p[2], lon, lat)[1])

        if mission == 'Sentinel-1b':
            if 75 < vaa < 105 and 20. < vza < 45.:  # vaa usually [0, 180], testing observation times
                geom_inst.date_utc.append(p[2])
                geom_inst.vza.append(vza)
                geom_inst.vaa.append(vaa)
                geom_inst.sza.append(sza)
                geom_inst.saa.append(saa)
        elif mission == 'Sentinel-1a':
            if 75 < vaa < 105 and 20. < vza < 45.:  # vaa usually [0, 180], testing observation times
                geom_inst.date_utc.append(p[2])
                geom_inst.vza.append(vza)
                geom_inst.vaa.append(vaa)
                geom_inst.sza.append(sza)
                geom_inst.saa.append(saa)
        elif mission == 'Sentinel-2a':
            if sza < 90 and vza < 10.3:
                geom_inst.date_utc.append(p[2])
                geom_inst.vza.append(vza)
                geom_inst.vaa.append(vaa)
                geom_inst.sza.append(sza)
                geom_inst.saa.append(saa)

    return geom_inst


def get_sentinel_geometry(startDateUTC, lengthDays, lat, lon, alt=0.0, mission="Sentinel-2a",
                          tleFile="/data/tle/norad_resource_tle.txt"):
    """Calculate approximate geometry for Sentinel overpasses.
    Approximate because it assumes maximum satellite elevation
    is the time at which target is imaged.

    :param startDateUTC: a datetime object specifying when to start prediction.
    :type startDateUTC: object
    :param lengthDays: number of days over which to perform calculations.
    :type lengthDays: int
    :param lat: latitude of target.
    :type lat: float
    :param lon: longitude of target.
    :type lon: float
    :param alt: altitude of target (in km).
    :type alt: float
    :param mission: mission name as in TLE file.
    :type mission: str
    :param tleFile: TLE file.
    :type tleFile: str
    :return: a python list containing instances of the sensorGeometry class arranged in date order.
    :rtype: list
    """
    orb = Orbital(mission, tleFile)
    passes = orb.get_next_passes(startDateUTC, 24 * lengthDays, lon, lat, alt)

    geomList = []

    for p in passes:
        look = orb.get_observer_look(p[2], lon, lat, alt)
        vza = 90 - look[1]
        vaa = look[0]
        sza = ast.sun_zenith_angle(p[2], lon, lat)
        saa = np.rad2deg(ast.get_alt_az(p[2], lon, lat)[1])

        if mission == 'Sentinel-1b':
            if 75 < vaa < 105 and 20. < vza < 45.:  # vaa usually [0, 180], testing observation times
                thisGeom = SensorGeometry()
                thisGeom.date_utc = p[2]
                thisGeom.vza = vza
                thisGeom.vaa = vaa
                thisGeom.sza = sza
                thisGeom.saa = saa
                geomList.append(thisGeom)
        elif mission == 'Sentinel-1a':
            if 75 < vaa < 105 and 20. < vza < 45.:  # vaa usually [0, 180], testing observation times
                thisGeom = SensorGeometry()
                thisGeom.date_utc = p[2]
                thisGeom.vza = vza
                thisGeom.vaa = vaa
                thisGeom.sza = sza
                thisGeom.saa = saa
                geomList.append(thisGeom)
        elif mission == 'Sentinel-2a':
            if sza < 90 and vza < 10.3:
                thisGeom = SensorGeometry()
                thisGeom.date_utc = p[2]
                thisGeom.vza = vza
                thisGeom.vaa = vaa
                thisGeom.sza = sza
                thisGeom.saa = saa
                geomList.append(thisGeom)

    return geomList


if __name__ == "__main__":
    """Simple test with S2a for Wallerfing
    """

    startDate = dt.datetime(2017, 7, 1, 0, 0, 0, 0, None)
    lon, lat = 12.880, 48.684
    alt = 0.
    days = 30

    geomList = get_sentinel_geometry(startDate, days, lat, lon, mission="Sentinel-1a", alt=alt)
    for g in geomList:
        g.print_geom()



"""A setuptools based setup module.
See:
https://packaging.python.org/en/latest/distributing.html
https://github.com/pypa/sampleproject
"""

# Always prefer setuptools over distutils
from setuptools import setup, find_packages
# To use a consistent encoding
from codecs import open
from os import path
import subprocess
from setuptools.command.install import install

here = path.abspath(path.dirname(__file__))

# Get the long description from the README file
with open(path.join(here, 'README.rst'), encoding='utf-8') as f:
    long_description = f.read()

class MyInstall(install):
    def run(self):
        try:
            # note cwd - this makes the current directory
            # the one with the Makefile.
            print path.join(here,'signaturesimulator/semidiscrete_srf/')
            subprocess.call(['make'], cwd=path.join(here, 'signaturesimulator/semidiscrete_srf/'))
        except Exception as e:
            print e
            print "Error compiling semiD.c.   Try running 'make'."
            exit(1)
        else:
            install.run(self)

# Arguments marked as "Required" below must be included for upload to PyPI.
# Fields marked as "Optional" may be commented out.

setup(
    # This is the name of your project. The first time you publish this
    # package, this name will be registered for you. It will determine how
    # users can install this project, e.g.:
    #
    # $ pip install sampleproject
    #
    # And where it will live on PyPI: https://pypi.org/project/sampleproject/
    #
    # There are some restrictions on what makes a valid project name
    # specification here:
    # https://packaging.python.org/specifications/core-metadata/#name
    name='signaturesimulator',  # Required

    # Versions should comply with PEP 440:
    # https://www.python.org/dev/peps/pep-0440/
    #
    # For a discussion on single-sourcing the version across setup.py and the
    # project code, see
    # https://packaging.python.org/en/latest/single_source_version.html
    version='0.2.2',  # Required

    # This is a one-line description or tagline of what your project does. This
    # corresponds to the "Summary" metadata field:
    # https://packaging.python.org/specifications/core-metadata/#summary
    description='A Python package that simulates satellite data from passive optical and active microwave sensors '
                'based on land surface biogeophysical state variables (leaf area index, canopy height and soil '
                'moisture) and viewing geometries and times.',  # Required

    # This is an optional longer description of your project that represents
    # the body of text which users will see when they visit PyPI.
    #
    # Often, this is the same as your README, so you can just read it in from
    # that file directly (as we have already done above)
    #
    # This field corresponds to the "Description" metadata field:
    # https://packaging.python.org/specifications/core-metadata/#description-optional
    long_description=long_description,  # Optional

    # This should be a valid link to your project's main homepage.
    #
    # This field corresponds to the "Home-Page" metadata field:
    # https://packaging.python.org/specifications/core-metadata/#home-page-optional
    url='https://github.com/pyearthsci/signaturesimulator',  # Optional

    # This should be your name or the name of the organization which owns the
    # project.
    author='T. Quaife and E. Pinnington',  # Optional

    # This should be a valid email address corresponding to the author listed
    # above.
    author_email='e.pinnington@reading.ac.uk',  # Optional

    # This field adds keywords for your project which will appear on the
    # project page. What does your project relate to?
    #
    # Note that this is a string of words separated by whitespace, not a list.
    keywords='satellite earth science land surface passive optical active microwave sentinel',  # Optional

    # You can just specify package directories manually here if your project is
    # simple. Or you can use find_packages().
    #
    # Alternatively, if you just want to distribute a single Python file, use
    # the `py_modules` argument instead as follows, which will expect a file
    # called `my_module.py` to exist:
    #
    #   py_modules=["my_module"],
    #
    packages=['signaturesimulator'],  # Required

    # This field lists other packages that your project depends on to run.
    # Any package you put here will be installed by pip when your project is
    # installed, so they must be valid existing projects.
    #
    # For an analysis of "install_requires" vs pip's requirements files see:
    # https://packaging.python.org/en/latest/requirements.html
    install_requires=['pandas',
                      'numpy',
                      'scipy',
                      'pyorbital',
                      'mock',
                      'matplotlib',
                      'netCDF4'],  # Optional

    # If there are data files included in your packages that need to be
    # installed, specify them here.
    #
    # If using Python 2.6 or earlier, then these have to be included in
    # MANIFEST.in as well.
    package_data={  # Optional
        'signaturesimulator': ['data/tle/*', 'data/srf/*', 'data/state_variables/*', 'data/geometries/*', 'sense/*',
                               'sense/dielectric/*', 'sense/surface/*', 'sense/vegetation/*', 'semidiscrete_srf/*']
    },

    cmdclass={'install': MyInstall},
)



# core module imports
import os
# third party imports
import numpy as np
import datetime as dt
import matplotlib.pyplot as plt
# signature simulator specific imports
import opticalcanopy_rt as opcan_rt
import activemicro_rt as actmicro_rt
import atmospheric_rt as atmo_rt
import satellite_geometry as satgeo
import state_vector as sv

# Define some example factory functions


def single_reflectance(vza=5.5, vaa=286.3, sza=26.8, saa=157.0, lai=3.0, canopy_ht=1.0, soil_m=0.3):
    """
    Factory function computing a single reflectance given viewing geometries of sensor and land surface state variables
    :param vza: view zenith angle (deg)
    :param vaa: view azimuth angle (deg)
    :param sza: solar zenith angle (deg)
    :param saa: solar azimuth angle (deg)
    :param lai: leaf area index (m2 m-2)
    :param canopy_ht: canopy height (m)
    :param soil_m: soil moisture (m3 m-3)
    :return: simulated Sentinel 2 reflectance for given input
    """
    s = Simulator()
    s.get_geom = s.geom_default(vza=vza, vaa=vaa, sza=sza, saa=saa)
    s.get_land_state = s.state_default(lai=lai, canopy_ht=canopy_ht, soil_m=soil_m)
    s.run_rt = s.passive_optical
    s.run()
    return s.spectra.refl


def multi_backscat(geom_csv='/data/geometries/s1_example_const.csv',
                   state_csv='/data/state_variables/state_example.csv'):
    """
    Factory function computing multiple values of backscatter given viewing geometries of sensor and land surface state
    variables specified in two csv files
    :param geom_csv: location of csv containing viewing geometries and times
    :param state_csv: location of csv containing land surface state variables and times
    :return: simulated Sentinel 1 backscatter in the hv polarisation
    """
    s = Simulator()
    s.get_geom = s.geom_csv(geom_csv)
    s.get_land_state = s.state_csv(state_csv)
    s.run_rt = s.active_microwave
    s.run()
    return s.backscat.hv


class Simulator(object):
    """Class to simulate satellite observations.
    """
    def __init__(self):
        """Calculate class attributes.
        """
        # find file path
        self.dir_path = os.path.dirname(os.path.realpath(__file__))
        # setup land surface state variables
        self.get_land_state = self.state_default()
        # setup geometry for observations
        self.get_geom = self.geom_default()
        # setup radiative transfer model
        self.run_rt = self.passive_optical  # choose from self.active_microwave or self.passive_optical
        # setup atmospheric RT model for passive optical sensors
        self.run_atmos_rt = atmo_rt.fwd_atm
        # setup keys for plotting variables
        self.output_variables = []
        # paths to example geometry and land state csv files
        self.example_s1_geometries = self.dir_path + '/' + 'data/geometries/s1_example_const.csv'
        self.example_s2_geometries = self.dir_path + '/' + 'data/geometries/s2_example_const.csv'
        self.example_state_variables = self.dir_path + '/' + 'data/state_variables/state_example.csv'

    def state_default(self, date_utc=dt.datetime(2016, 6, 17, 9, 0), lai=3.0, canopy_ht=1.0, soil_m=0.3):
        """
        Defines a default land state given some input
        :param date_utc: date as a datetime object
        :type date_utc: object
        :param lai: Leaf area index at specified time (m2 m-2)
        :type lai: float
        :param canopy_ht: canopy height at specified time (m)
        :type canopy_ht: float
        :param soil_m: soil moisture at specified time (m3 m-3)
        :type soil_m: float
        :return: instance of StateVector class
        """
        state_inst = sv.StateVector()
        state_inst.date_utc = [date_utc]
        state_inst.lai = [lai]
        state_inst.can_height = [canopy_ht]
        state_inst.soil_moisture = [soil_m]
        return state_inst

    def state_csv(self, fname):
        """
        Extract data from a .csv file and create a state instance.
        :param fname: Filename and location of .csv to extract data from
        :type fname: str
        :return: instance of StateVector class
        """
        state_inst = sv.get_state_csv(fname)
        return state_inst

    def state_jules(self, land_cover='crop'):
        """Extracts state variables from either a) a specified JULES netcdf file, or b) a default land cover profile
        from the list: ['crop', 'evergreen', 'broadleaf', 'grassland']
        :param land_cover: Either a filepath to JULES output netcdf file or a string from list:
        ['crop', 'evergreen', 'broadleaf', 'grassland']
        :type land_cover: str
        :return: instance of StateVector class
        """
        if land_cover in ['crop', 'crop_early', 'crop_late']:
            state_inst = sv.get_jules_state(self.dir_path+'/'+'../data/state_variables/'+land_cover+'.nc')
        else:
            state_inst = sv.get_jules_state(land_cover)
        return state_inst

    def geom_default(self, date_utc=dt.datetime(2016, 6, 17, 10, 25), vza=5.5, vaa=286.3, sza=26.8, saa=157.0):
        """
        Defines geometry for satellite observation
        :param date_utc: date of observation as a datetime object
        :type date_utc: object
        :param vza: view zenith angle (degrees)
        :type vza: float
        :param vaa: view azimuth angle (degrees)
        :type vaa: float
        :param sza: solar zenith angle (degrees)
        :type sza: float
        :param saa: solar azimuth angle (degrees)
        :type saa: float
        :return: instance of SensorGeometry class
        """
        geom_inst = satgeo.SensorGeometry()
        geom_inst.date_utc = [date_utc]
        geom_inst.vza = [vza]
        geom_inst.vaa = [vaa]
        geom_inst.sza = [sza]
        geom_inst.saa = [saa]
        return geom_inst

    def geom_csv(self, fname):
        """
        Extracts dates and viewing geomteries from a .csv and creates an instance of the SensorGeometry class
        :param fname: Filename and location of .cvs file containing dates and viewing angles for sensor
        :type fname: str
        :return: instance of SensorGeometry class
        """
        geom_inst = satgeo.get_geom_csv(fname)
        return geom_inst

    def geom_pyoribtal(self, start_date=dt.datetime(2016, 1, 1), num_days=365, lon=12.880, lat=48.684, altitude=0.0,
                       sat_mission='S2', tle_file=None):
        """
        Uses pyorbital to calculate sensor geometries for a specified time period and location on the earth
        :param start_date: start date for period in which to get sensor geometries
        :type start_date: datetime object
        :param num_days: number of days in period to get sensor geometries
        :type num_days: int
        :param lon: location longitude
        :type lon: float
        :param lat: location latitude
        :type lat: float
        :param altitude: location altitude
        :type altitude: float
        :param sat_mission: satellite mission
        :type sat_mission: str
        :param tle_file: tle file location with parameters for given sensor satellite mission
        :type tle_file: str
        :return: instance of SensorGeometry class
        """
        mission_dict = {'S2': 'Sentinel-2a', 'S1': 'Sentinel-1b'}
        if tle_file is None:
            tle_file = self.dir_path + "/data/tle/norad_resource_tle.txt"
        geom_inst = satgeo.get_satellite_geometry(start_date, num_days, lon, lat, alt=altitude,
                                                  mission=mission_dict[sat_mission], tleFile=tle_file)
        return geom_inst

    def passive_optical(self, state, geom,):
        """
        Function that runs semi discrete rt model for passive optical sensors and returns instance of the spectra class
        :param state: instance of the StateVector class
        :type state: object
        :param geom: instance of the SensorGeometry class
        :type geom: object
        :return: instance of the spectra class
        """
        spectra = opcan_rt.passive_optical_rt(state, geom)
        return spectra

    def active_microwave(self, state, geom,):
        """
        Function that runs sense model for active microwave sensors and returns instance of the backscatter class
        :param state: instance of the StateVector class
        :type state: object
        :param geom: instance of the SensorGeometry class
        :type geom: object
        :return: instance of the backscatter class
        """
        backscat = actmicro_rt.active_microwave_rt(state, geom)
        return backscat

    def run(self):
        """
        Class method that runs the simulator for specifed viewing geometries and land surface state variables
        """
        if self.run_rt == self.passive_optical:
            self.spectra = self.run_rt(self.get_land_state, self.get_geom)
            self.spectra = self.run_atmos_rt(self.spectra)
            self.output_variables += self.spectra.__dict__.keys()
            self.output_variables.remove('ftype')
            self.output_variables.remove('wavl')
        elif self.run_rt == self.active_microwave:
            self.backscat = self.run_rt(self.get_land_state, self.get_geom)
            self.output_variables += self.backscat.__dict__.keys()

    def plot(self, plot_key, band_idx=None):
        """
        Function to plot output of simulator after completing a run
        :param plot_key: plot key of variable to plot (must be in self.plot_keys list)
        :type plot_key: str
        :param band_idx: index to reflectance band to plot if plot key "refl" is specified
        :type band_idx: int
        :return: requested plot
        """
        try:
            if plot_key not in self.output_variables:
                return "Plot key is not in specified plot keys list for simulator output"
            elif plot_key == "refl" and band_idx == None:
                return "For plot key refl you must also specify a band index using, band_idx=..."
            elif plot_key == "refl":
                plt.plot(self.spectra.date_sat_ob, np.array(self.spectra.__dict__["refl"])[:, band_idx], 'o')
                plt.xlabel('Date')
                plt.ylabel('Band reflectance')
                plt.show()
            elif self.run_rt == self.active_microwave:
                plt.plot(self.backscat.date_sat_ob, self.backscat.__dict__[plot_key], 'o')
                plt.xlabel('Date')
                plt.ylabel(plot_key)
                plt.show()
            elif self.run_rt == self.passive_optical:
                plt.plot(self.spectra.date_sat_ob, self.spectra.__dict__[plot_key], 'o')
                plt.xlabel('Date')
                plt.ylabel(plot_key)
                plt.show()
            else:
                return "Something went wrong, please check that you have run the simulator and are specifying a valid" \
                       "plot key"
        except AttributeError:
            print('Please run simulator before plotting, using e.g. simulator_instance.run()')


"""
Major surface scatter class
"""
class SurfaceScatter(object):
    def __init__(self, eps, ks, theta, kl=None, **kwargs):
        self.eps = eps
        self.ks = ks
        self.theta = theta
        self.kl = kl
        self._check()

    def _check(self):
        assert isinstance(self.eps, complex)

class SurfaceScatterWaterCloudModel(object):
    def __init__(self, mv, theta, C=None, C1=None, C2=None, D=None):
        self.mv = mv
        self.theta = theta
        self.C = C
        self.C1 = C1
        self.C2 = C2
        self.D = D






"""
Specification of canopies
"""

class Canopy(object):
    def __init__(self, **kwargs):
        self.d = kwargs.get('d', None)
        self._check()

    def _check(self):
        assert self.d is not None, 'Vegetation height needs to be given'



class OneLayer(Canopy):
    """
    define a homogeneous one layer canopy
    """
    def __init__(self, **kwargs):
        super(OneLayer, self).__init__(**kwargs)
        self.ke_h = kwargs.get('ke_h', None)
        self.ke_v = kwargs.get('ke_v', None)
        assert self.ke_h is not None
        assert self.ke_v is not None
        self.ks_h = kwargs.get('ks_h', None)
        self.ks_v = kwargs.get('ks_v', None)
        assert self.ks_h is not None
        assert self.ks_v is not None




#!/usr/bin/env python

import re
from copy import copy
import numpy as np


class UnknownFileType(Exception):
    """Exception class for unknown filetypes
    """
    pass


class spectra(object):
    """Class holding spectra information for signaturesimulator.
    """
    def __init__(self, fname=None, ftype="SVC", wavlCol=0, reflCol=1, hdrLines=1):

        self.refl = np.array([])
        self.wavl = np.array([])
        self.ftype = ftype

        if fname != None:
            self.loadSpectra(fname, wavlCol, reflCol, hdrLines)

    def loadSpectra(self, fname, wavlCol=0, reflCol=1, hdrLines=1):
        """Load in the spectra from a given file using a
        method appropriate to the type of file.

        .. note:: Current supported formats are:

         SVC - SCV .sig ascii file

         CSV - standard ascii comma seperated values

        :param fname: Valid filename containing spectra.
        :type fname: str
        :param wavlCol: Column containing wavelengths.
        :type wavlCol: int
        :param reflCol: Column containing reflectance data.
        :type reflCol: int
        :param hdrLines: Number of lines to skip at start of file.
        :type hdrLines: int
        """

        with open(fname) as f:
            if self.ftype == "SVC":
                self.loadSVCSig(f)
            elif self.ftype == "CSV":
                self.loadCSV(f, wavlCol, reflCol, hdrLines)
            else:
                raise UnknownFileType(self.ftype)
            f.close()

    def loadCSV(self, f, wavlCol=0, reflCol=1, hdrLines=1):
        """Read in data from a standard CSV file object.

        :param f: File object.
        :type f: file
        :param wavlCol: Column containing wavelengths.
        :type wavlCol: int
        :param reflCol: Column containing reflectance data.
        :type reflCol: int
        :param hdrLines: Number of lines to skip at start of file.
        :type hdrLines: int
        """

        tmp = np.loadtxt(f, delimiter=",", skiprows=hdrLines, usecols=(wavlCol, reflCol))
        self.wavl = tmp[:, 0]
        self.refl = tmp[:, 1]

    def loadSVCSig(self, f):
        """Read in data from an SVC .sig ascii file.

        :param f: File object.
        :type f: file
        """

        getData = False

        for line in f:
            if getData:
                reflTmp = np.append(self.refl, float(line.split()[3]) / 100.)
                wavlTmp = np.append(self.wavl, float(line.split()[0]))
                self.refl = copy(reflTmp)
                self.wavl = copy(wavlTmp)
            if re.match('data=', line):
                getData = True

    def interpolate(self, resltn=0.1):
        """Interpolate spectra to the given resolution.
        Overwites exisiting data.

        :param resltn: resolution of the interpolation.
        :type resltn: float
        """

        # find the starting and ending wavelengths
        begWavl = np.ceil(self.wavl[0] / resltn) * resltn
        endWavl = np.floor(self.wavl[-1] / resltn) * resltn

        # print self.wavl[0], begWavl
        # print self.wavl[-1], endWavl

        # generate new wavelength and relfectance arrays
        wavlTmp = np.arange(begWavl, endWavl + resltn, resltn)
        reflTmp = np.zeros(np.shape(wavlTmp))

        # perfrom a linear interpolation:
        m = 0
        for (n, wavl) in enumerate(wavlTmp):

            while self.wavl[m] < wavl and wavl != wavlTmp[-1]:
                m += 1

            if self.wavl[m] == wavl:
                reflTmp[n] = self.refl[m]
            else:
                w1 = self.wavl[m - 1]
                w2 = self.wavl[m]
                r1 = self.refl[m - 1]
                r2 = self.refl[m]
                f = (w2 - wavl) / (w2 - w1)
                reflTmp[n] = r1 * f + r2 * (1 - f)

        # copy in interploated data
        self.wavl = copy(wavlTmp)
        self.refl = copy(reflTmp)

    def trim(self, wlmin, wlmax):
        """Trim the spectra so it is between two specified wavelengths. Destroys the original data.

        :param wlmin: The lowest wavelength of the new spectra.
        :type wlmin: float
        :param wlmax: The highest wavelength of the new spectra.
        :type wlmax: float
        """

        reflTmp = np.array([])
        wavlTmp = np.array([])

        for (n, wavl) in enumerate(self.wavl):
            if (wavl >= wlmin - 1e-09) and (wavl <= wlmax + 1e-09):
                wavlTmp2 = np.append(wavlTmp, self.wavl[n])
                reflTmp2 = np.append(reflTmp, self.refl[n])
                wavlTmp = copy(wavlTmp2)
                reflTmp = copy(reflTmp2)
        # copy over trimmed data
        self.wavl = copy(wavlTmp)
        self.refl = copy(reflTmp)


def convolve(s1orig, s2orig, resln=1.0, s2norm=True):
    """Convolve one spectra with another, for example
    to apply a band pass, or a spectral response function.

    :param s1orig: A spectra object.
    :type s1orig: object
    :param s2orig: A spectra object.
    :type s2orig: float
    :param resln: The spectral resolution to use.
    :type resln: float
    :param s2norm: If True normalise the second spectra (e.g. to apply a spectra response function).
    :type s2norm: bool
    :return: Convolved spectra.
    :rtype: object
    """

    # make copies so as not to alter
    # original data
    s1 = copy(s1orig)
    s2 = copy(s2orig)

    # interpolate to common resolution
    s1.interpolate(resln)
    s2.interpolate(resln)

    # trim spectra to encompass the exclusive
    # range of the two
    wlmin = np.max([s1.wavl[0], s2.wavl[0]])
    wlmax = np.min([s1.wavl[-1], s2.wavl[-1]])
    s1.trim(wlmin, wlmax)
    s2.trim(wlmin, wlmax)

    # convolve and normailse if required
    norm = 1.0
    if s2norm:
        norm = s2.refl.sum()
    return np.dot(s1.refl, s2.refl) / norm


def sentinel2(s, mission="a"):
    srf = []
    sen2 = spectra()

    if mission == "a":
        fname = "../srfData/S2a_SRF.csv"
    else:
        fname = "../srfData/S2b_SRF.csv"

    for n in xrange(1, 14):
        srf = spectra(fname=fname, ftype="CSV", reflCol=n)
        reflTmp = np.append(sen2.refl, convolve(s, srf, 0.1))
        sen2.refl = copy(reflTmp)
        m = np.argmax(srf.refl)
        wavlTmp = np.append(sen2.wavl, srf.wavl[m])
        sen2.wavl = copy(wavlTmp)

    return sen2


if __name__ == "__main__":

    from matplotlib import pyplot as plt

    doTest1 = True
    doTest2 = False

    if doTest1:
        # test simulation of S2 bands
        svc = spectra(fname="../testData/HRPDA.053017.0065_moc.sig")
        sen2a = sentinel2(svc)
        sen2b = sentinel2(svc, mission="b")

        plt.plot(svc.wavl, svc.refl)
        plt.plot(sen2a.wavl, sen2a.refl, 'o-')
        plt.plot(sen2b.wavl, sen2b.refl, 'o-')
        plt.xlabel('wavelength (nm)')
        plt.ylabel('relfectance (-)')
        plt.show()

    if doTest2:
        # test interpolation and trim routines
        s = spectra(fname="testData/HRPDA.053017.0065_moc.sig")
        plt.plot(s.wavl, s.refl)
        s.interpolate(50)
        plt.plot(s.wavl, s.refl, 'o')
        s.interpolate(100)
        s.trim(1200, 2000)
        plt.plot(s.wavl, s.refl, '--')
        plt.show()
